import sys, os
import copy as cp
import numpy as np
from pmx import *
from pmx import geometry
from pmx import ndx
from pmx import library
from pmx.forcefield import *
import scipy
from scipy.special import erf
from string import digits
from pmx import library as pmxlib

class AbsRestraints:

    def __init__( self, **kwargs ):

        # constants
        R = 8.31445985*0.001  # Gas constant in kJ/mol/K
        self.T = 298.15
        self.RT = R*self.T

        # structure: one of those need to be present
        self.system = None # pmx model
        self.strFile = None 

        # booleans
        self.bLigMaxSpread = False
        self.bAnyHeavyAtom = False # set this to true to select any heavy atom on protein's side
        self.bManLig = False
        self.bManProt = False
        self.bVerbose = False

        # force constants
        self.kBond = 4184.0
        self.kAngle = 41.84
        self.kDihedral = 41.84

        # files
        self.topLigFile = None
        self.topProtFile = None
        self.ndxFile = None
        self.dgFile = None
        self.outiiFile = None
        self.pdbRestrAtoms = None # an optional output with the restrained atoms only (convenient for viewing)

        # ligandID: may provide manually
        self.indLig1 = False
        self.indLig2 = False
        self.indLig3 = False

        # proteinID: may provide manually
        self.indProt1 = False
        self.indProt2 = False
        self.indProt3 = False

        # center ligand name
        self.ligCenterByName = False

        # indeces
        self.indLig = []
        self.indProt = []

        for key, val in kwargs.items():
            setattr(self,key,val)

        # pdb file is necessary
        if self.strFile==None:
            sys.stdout.write('Need to provide structure file')
            sys.exit(0)
        self.system = Model(self.strFile) 

        # topologies
        self.protTop = None
        self.ligTop = None
        if len(self.topLigFile)!=0:
            self.ligTop = TopolBase( self.topLigFile )
        if len(self.topProtFile)!=0:
            self.protTop = TopolBase( self.topProtFile )

        # merge topologies
        # if self.ligTop != None and self.protTop != None:
#            mergedTop = self._merge_top( )
#            renumLigTop = self._renum_lig_top( )

        if len(self.indProt)==0:
            self.indProt = self._select_ndx(self.ndxFile,message='Select index group for protein:\n')
        if len(self.indLig)==0:
            self.indLig = self._select_ndx(self.ndxFile,message='Select index group for ligand:\n')

        # ligand indeces
        if self.bManLig==True:
            if self.indLig1==False:
                self.indLig1 = self._select_ndx(self.ndxFile,message='Select first ligand atom for distance restraint:\n')
                self._check_ndx_size( len(self.indLig1), desiredSize=1 )
            if self.indLig2==False:
                self.indLig2 = self._select_ndx(self.ndxFile,message='Select second ligand atom for angle restraint:\n')
                self._check_ndx_size( len(self.indLig2), desiredSize=1 )
            if self.indLig3==False:
                self.indLig3 = self._select_ndx(self.ndxFile,message='Select third ligand atom for dihedral restraint:\n')
                self._check_ndx_size( len(self.indLig3), desiredSize=1 )
        elif self.ligCenterByName!=False: # central ligand atom provided manually, the rest of atoms automatically
            self.indLig1 = self._ligAtomByName( self.ligCenterByName, self.system )
            if self.indLig1==False:
                sys.stdout.write('Cannot find the ligand atom ',self.ligCenterByName)
                sys.stdout.write('Exiting...')
                sys.exit(1)
#            indLig1 = self._select_ndx(ndxFile,message='Select one ligand atom for distance restraint:\n')
#            self._check_ndx_size( len(indLig1), desiredSize=1 )
            (self.indLig1,self.indLig2,self.indLig3) = self._identify_ligand_atoms( self.system, self.indLig1 )
        else: # find all ligand atoms automatically
            (self.indLig1,self.indLig2,self.indLig3) = self._identify_ligand_atoms( self.system, self.indLig1 )

        # atoms
        self.aLig1 = self.system.atoms[self.indLig1[0]]
        self.aLig2 = self.system.atoms[self.indLig2[0]]
        self.aLig3 = self.system.atoms[self.indLig3[0]]
        self.aLig1.a2nm()
        self.aLig2.a2nm()
        self.aLig3.a2nm()

        # protein indeces
        if self.bManProt==True:
            if self.indProt1==False:
                self.indProt1 = self._select_ndx(self.ndxFile,message='Select first protein atom for distance restraint:\n')
                self._check_ndx_size( len(self.indProt1), desiredSize=1 )
            if self.indProt2==False:
                self.indProt2 = self._select_ndx(self.ndxFile,message='Select second protein atom for angle restraint:\n')
                self._check_ndx_size( len(self.indProt2), desiredSize=1 )
            if self.indProt3==False:
                self.indProt3 = self._select_ndx(self.ndxFile,message='Select third protein atom for dihedral restraint:\n')
                self._check_ndx_size( len(self.indProt3), desiredSize=1 )
        else: # find protein indeces
            bRandProtAtoms = False
            if bRandProtAtoms==True:
                (self.indProt1,self.indProt2,self.indProt3) = self._random_protein_atoms( self.system )
            else:
                (self.indProt1,self.indProt2,self.indProt3) = self._identify_protein_atoms( self.system, self.indLig1, self.indLig2, self.indLig3 )
        # atoms
        self.aProt1 = self.system.atoms[self.indProt1[0]]
        self.aProt2 = self.system.atoms[self.indProt2[0]]
        self.aProt3 = self.system.atoms[self.indProt3[0]]
        self.aProt1.a2nm()
        self.aProt2.a2nm()
        self.aProt3.a2nm()

        # distance
        self.dist = self.aProt1 - self.aLig1

        # angles
        # aLig2 - aLig1 - aProt1
        self.ang1 = self.aLig1.angle(self.aLig2,self.aProt1,degree=True)
        self._check_angle( self.ang1 )
        # aLig1 - aProt1 - aProt2
        self.ang2 = self.aProt1.angle(self.aLig1,self.aProt2,degree=True)
        self._check_angle( self.ang2 )

        # dihedrals
        # aLig3 - aLig2 - aLig1 - aProt1
        self.dih1 = self.aLig3.dihedral(self.aLig2,self.aLig1,self.aProt1,degree=True)
        # aLig2 - aLig1 - aProt1 - aProt2
        self.dih2 = self.aLig2.dihedral(self.aLig1,self.aProt1,self.aProt2,degree=True)
        # aLig1 - aProt1 - aProt2 - aProt3
        self.dih3 = self.aLig1.dihedral(self.aProt1,self.aProt2,self.aProt3,degree=True)


        # generate [ intermolecular_interactions ]
#        mergedTop.has_ii = True
        self.ii = {}
        self._gen_ii_bonds()
        self._gen_ii_angles()
        self._gen_ii_dihedrals()

        # optional output of restrained atoms as .pdb
        if self.pdbRestrAtoms!=None:
            mout = Model()
            mout.atoms.extend( (self.aLig1,self.aLig2,self.aLig3,self.aProt1,self.aProt2,self.aProt3) )
            mout.write(self.pdbRestrAtoms)

        # write intermolecular interaction
        self._write_ii( self.outiiFile )

        # calculate restraint contribution to the free energy
        if self.dgFile!=None: 
            fp = open(self.dgFile,'w')
        self._calc_restraint_dg( )
        if self.bVerbose==True:
            sys.stdout.write('Restraint contribution to free energy (according to B-K): %3.4f kJ/mol\n' % self.dg )
            sys.stdout.write('Restraint contribution to free energy (according to B-K): %3.4f kcal/mol\n' % self.dgkcal )
            sys.stdout.write('\n')
        if self.dgFile!=None: 
            fp.write('Restraint contribution to free energy (according to B-K): %3.4f kJ/mol\n' % self.dg )
            fp.write('Restraint contribution to free energy (according to B-K): %3.4f kcal/mol\n' % self.dgkcal )
            fp.write('\n')

        self._calc_restraint_dg_w_limits( )
        if self.bVerbose==True:
            sys.stdout.write('Restraint contribution to free energy (w ideal limits): %3.4f kJ/mol\n' % self.dg )
            sys.stdout.write('Restraint contribution to free energy (w ideal limits): %3.4f kcal/mol\n' % self.dgkcal )
            sys.stdout.write('\n')
        if self.dgFile!=None: 
            fp.write('Restraint contribution to free energy (w ideal limits): %3.4f kJ/mol\n' % self.dg )
            fp.write('Restraint contribution to free energy (w ideal limits): %3.4f kcal/mol\n' % self.dgkcal )
            fp.write('\n')
        
        self._calc_restraint_dg_w_gromacs_limits( )
        if self.bVerbose==True:
            sys.stdout.write('Restraint contribution to free energy (w gmx limits): %3.4f kJ/mol\n' % self.dg )
            sys.stdout.write('Restraint contribution to free energy (w gmx limits): %3.4f kcal/mol\n' % self.dgkcal )
        if self.dgFile!=None: 
            fp.write('Restraint contribution to free energy (w gmx limits): %3.4f kJ/mol\n' % self.dg )
            fp.write('Restraint contribution to free energy (w gmx limits): %3.4f kcal/mol\n' % self.dgkcal )
            fp.close()

        # molecule name
#        mergedTop.name = 'merged'

        # write topology
 #       if outItpFile!=False:
  #          renumLigTop.write(outItpFile,stateBonded='A')
        #    mergedTop.write(outItpFile,stateBonded='A')

    def _check_angle(self, angle):
        extr1 = 0.5*self.kAngle*np.power((angle-0.0)/180.0*np.pi,2)
        extr2 = 0.5*self.kAngle*np.power((angle-180.0)/180.0*np.pi,2)
        if extr1/self.RT<5.0:
            if self.bVerbose==True:
                sys.stdout.write('Warning: one of the angle restraints is less than 5 kT away from 0 degree angle\n')
            return(False)
        elif extr2/self.RT<5.0:
            if self.bVerbose==True:
                sys.stdout.write('Warning: one of the angle restraints is less than 5 kT away from 180 degree angle\n')
            return(False)
        return(True)

    def _calc_restraint_dg(self):
        """Calculates effect of restraints in gas phase following https://doi.org/10.1021/jp0217839.
        
        This is the approximate result with no limits placed on the restraint coordinates.
        """
        
        V0 = 1.66            # standard volume in nm^3
        dgPrefactor = ( 8.0*np.power(np.pi,2.0)*V0/(np.power(self.dist,2.0)*np.sin(self.ang1*np.pi/180.0)*np.sin(self.ang2*np.pi/180.0)) )
        dgForceConstants = np.sqrt(self.kBond*self.kAngle*self.kAngle*self.kDihedral*self.kDihedral*self.kDihedral)/np.power(2.0*np.pi*self.RT,3.0)
        self.dg = -self.RT * np.log(dgPrefactor*dgForceConstants)
        self.dgkcal = self.dg/4.184
        
        
    def _calc_restraint_dg_w_limits(self):
        """Calculates effect of restraints in gas phase taking limits into account.
        
        This is the "ideal" case with limits symmetric around the equilibrium angles.
        dist       [0, inf]
        ang-ang_0  [-Pi, Pi]
        dih-dih_0  [-Pi, Pi]
        """
        
        V0 = 1.66            # standard volume in nm^3
        
        #-1^6 = 1;
        #Boresch, ... and Karplus put their Is in here
        # and they work out to I=-2 given infinite limits of constrant variables.
        # So in their case there is a sqrt(2*Pi*RT)^6 instead of sqrt(0.5*Pi*RT)^6
        integral_multiplier = np.power(np.pi*self.RT/2.0, 3.0)
        
        I_dist = -1 - erf( np.sqrt(self.kBond/(2.0*self.RT)) * self.dist )
        I_ang = erf( np.sqrt(self.kAngle/(2.0*self.RT)) * (-np.pi) ) - \
                erf( np.sqrt(self.kAngle/(2.0*self.RT)) * (np.pi) )
        I_dih = erf( np.sqrt(self.kDihedral/(2.0*self.RT)) * (-np.pi) ) - \
                erf( np.sqrt(self.kDihedral/(2.0*self.RT)) * (np.pi) )
        
        I_ang = [I_ang, I_ang]
        I_dih = [I_dih, I_dih, I_dih]
        
        forceConstants = np.sqrt(self.kBond*self.kAngle*self.kAngle*self.kDihedral*self.kDihedral*self.kDihedral)
        partition_func = np.power(self.dist,2.0)*np.sin(self.ang1*np.pi/180.0)*np.sin(self.ang2*np.pi/180.0) * \
                         integral_multiplier * I_dist * I_ang[0] * I_ang[1] * \
                         I_dih[0] * I_dih[1] * I_dih[2] / forceConstants
        
        self.dg = -self.RT * np.log(8.0*np.power(np.pi,2.0)*V0 / partition_func)
        self.dgkcal = self.dg/4.184
        
        
    def _calc_restraint_dg_w_gromacs_limits(self):
        """Calculates effect of restraints in gas phase using limits of gromacs harmonic restraints.
        
        This is the case for harmonic restraint potentials in gromacs.
        dist       [0, inf]
        ang        [0, Pi]
        dih-dih_0  [-Pi, Pi]
        """
        
        V0 = 1.66            # standard volume in nm^3
        
        #-1^6 = 1;
        #Boresch, ... and Karplus put their Is in here
        # and they work out to I=-2 given infinite limits of constrant variables.
        # So in their case there is a sqrt(2*Pi*RT)^6 instead of sqrt(0.5*Pi*RT)^6
        integral_multiplier = np.power(np.pi*self.RT/2.0, 3.0)
        
        I_dist = -1 - erf( np.sqrt(self.kBond/(2.0*self.RT)) * self.dist )
        I_angA = erf( np.sqrt(self.kAngle/(2.0*self.RT)) * ((self.ang1*np.pi/180.0)-np.pi) ) - \
                 erf( np.sqrt(self.kAngle/(2.0*self.RT)) * (self.ang1*np.pi/180.0) )
        I_angB = erf( np.sqrt(self.kAngle/(2.0*self.RT)) * ((self.ang2*np.pi/180.0)-np.pi) ) - \
                 erf( np.sqrt(self.kAngle/(2.0*self.RT)) * ((self.ang2*np.pi/180.0)) )
        I_dih = erf( np.sqrt(self.kDihedral/(2.0*self.RT)) * (-np.pi) ) - \
                erf( np.sqrt(self.kDihedral/(2.0*self.RT)) * (np.pi) )
                
        #print I_angA, I_angB
#        print erf( sqrt(self.kAngle/(2.0*self.RT)) * (self.ang1-np.pi) ), sqrt(self.kAngle/(2.0*self.RT)) * (self.ang1-np.pi), \
#              erf( sqrt(self.kAngle/(2.0*self.RT)) * (self.ang1) ), sqrt(self.kAngle/(2.0*self.RT)) * (self.ang1)
                
        I_ang = [I_angA, I_angB]
        I_dih = [I_dih, I_dih, I_dih]
        
        forceConstants = np.sqrt(self.kBond*self.kAngle*self.kAngle*self.kDihedral*self.kDihedral*self.kDihedral)
        partition_func = np.power(self.dist,2.0)*np.sin(self.ang1*np.pi/180.0)*np.sin(self.ang2*np.pi/180.0) * \
                         integral_multiplier * I_dist * I_ang[0] * I_ang[1] * \
                         I_dih[0] * I_dih[1] * I_dih[2] / forceConstants
        
        self.dg = -self.RT * np.log(8.0*np.power(np.pi,2.0)*V0 / partition_func)
        self.dgkcal = self.dg/4.184

    def _write_ii( self, outiiFile ):
        fp = open(outiiFile, 'w')
        fp.write('\n [ intermolecular_interactions ]\n')
        # bonds
        if 'bonds' in self.ii.keys():
           fp.write(' [ bonds ]\n')
           for b in self.ii['bonds']:
               fp.write('%6d %6d %6d' % ( b[0].id, b[1].id, b[2] ))
               if len(b)>3:
                   for x in b[3]:
                       fp.write(' %14.6f' % x)
               fp.write('\n')
           fp.write('\n')
        # angles
        if 'angles' in self.ii.keys():
           fp.write(' [ angles ]\n')
           for ang in self.ii['angles']:
               fp.write('%6d %6d %6d %6d' % ( ang[0].id, ang[1].id, ang[2].id, ang[3] ))
               if len(ang)>4:
                   for x in ang[4]:
                       fp.write(' %14.6f' % x)
               fp.write('\n')
           fp.write('\n')
        # dihedrals
        if 'dihedrals' in self.ii.keys():
           fp.write(' [ dihedrals ]\n')
           for dih in self.ii['dihedrals']:
               fp.write('%6d %6d %6d %6d %6d' % ( dih[0].id, dih[1].id, dih[2].id, dih[3].id, dih[4] ))
               if len(dih)>5:
                   for x in dih[5]:
                       fp.write(' %14.6f' % x)
               fp.write('\n')
           fp.write('\n')
        fp.close()

    def _gen_ii_bonds( self ):
        bond1 = [ self.aLig1, self.aProt1, 6, [self.dist, 0.0, self.dist, self.kBond] ]
        self.ii['bonds'] = [bond1]

    def _gen_ii_angles( self ):
        angle1 = [ self.aLig2, self.aLig1, self.aProt1, 1, [self.ang1, 0.0, self.ang1, self.kAngle] ]
        angle2 = [ self.aLig1, self.aProt1, self.aProt2, 1, [self.ang2, 0.0, self.ang2, self.kAngle] ]
        self.ii['angles'] = [ angle1, angle2 ]

    def _gen_ii_dihedrals( self ):
        dihedral1 = [ self.aLig3, self.aLig2, self.aLig1, self.aProt1, 2, [self.dih1, 0.0, self.dih1, self.kDihedral] ]
        dihedral2 = [ self.aLig2, self.aLig1, self.aProt1, self.aProt2, 2, [self.dih2, 0.0, self.dih2, self.kDihedral] ]
        dihedral3 = [ self.aLig1, self.aProt1, self.aProt2, self.aProt3, 2, [self.dih3, 0.0, self.dih3, self.kDihedral] ]
        self.ii['dihedrals'] = [ dihedral1, dihedral2, dihedral3 ]

    def _renum_lig_top( self ):
        renumTop = cp.deepcopy(self.ligTop)
        protAtomNum = len(self.protTop.atoms)
        # atoms
        for a in renumTop.atoms:
            a.id += protAtomNum
            a.cgnr += protAtomNum
        return(renumTop)

    def _merge_top( self ):
        mergedTop = cp.deepcopy(self.protTop)
        protAtomNum = len(mergedTop.atoms)
        # atomtypes
        if self.ligTop.has_atomtypes==True:
            mergedTop.has_atomtypes=True
            for atype in self.ligTop.atomtypes:
                mergedTop.atomtypes.append(atype)
        if self.protTop.has_atomtypes==True:
            mergedTop.has_atomtypes=True
            for atype in self.protTop.atomtypes:
                mergedTop.atomtypes.append(atype)
        # atoms
        for a in self.ligTop.atoms:
            a.id += protAtomNum
            a.cgnr += protAtomNum
            mergedTop.atoms.append(a)
        # bonds
        for b in self.ligTop.bonds:
            mergedTop.bonds.append(b)
        # pairs
        for p in self.ligTop.pairs:
            mergedTop.pairs.append(p)
        # angles
        for ang in self.ligTop.angles:
            mergedTop.angles.append(ang)
        # dihedrals
        for dih in self.ligTop.dihedrals:
            mergedTop.dihedrals.append(dih)

        return(mergedTop)

    def _change_outfile_format(filename, ext):
        head, tail = os.path.split(filename)
        name, ex = os.path.splitext(tail)
        new_name = os.path.join(head, name+'.'+ext)
        return new_name

    def _random_protein_atoms( self, system ):
        n = len(system.atoms)
        rand = np.random.choice(n, size=3, replace=False)
        at = system.atoms
        a1 = at[rand[0]]
        a2 = at[rand[1]]
        a3 = at[rand[2]]
 
        indProt1 = a1.id-1
        indProt2 = a2.id-1
        indProt3 = a3.id-1

        return([indProt1],[indProt2],[indProt3])

    def _get_mass( self, aname ):
        aname = aname.translate(str.maketrans('','',digits))
        if aname.startswith('BR') or aname.startswith('Br') or aname.startswith('br'):
            return(pmxlib._atommass['BR'])
        elif aname.startswith('CL') or aname.startswith('Cl') or aname.startswith('cl'):
            return(pmxlib._atommass['CL'])
#        elif aname.startswith('D'):
#            return(pmxlib._atommass[aname[1]])
        elif aname.startswith('H'):
            return(0.0)
        elif aname.startswith('E'):
            return(0.0)
        else:
            try:
                return(pmxlib._atommass[aname[0]])
            except:
                return(1.0)
        return(1.0)

    def _get_atom_name( self, aname ):
        if self.bAnyHeavyAtom==True:
            return('C')
        aname = aname.translate(str.maketrans('','',digits))
        if ('CL' in aname) or ('Cl' in aname) or ('cl' in aname):
            return('CL')
        if ('BR' in aname) or ('Br' in aname) or ('br' in aname):
            return('BR')
        return(aname[0])

    def _get_lig_COM( self, resname, system ):
        com = [0.0, 0.0, 0.0]
        mass = 0.0
        for a in system.atoms:
            if a.resname==resname and (a.id-1 in self.indLig):
                a.a2nm()
                amass = self._get_mass( a.name )
                com[0] += a.x[0]*amass
                com[1] += a.x[1]*amass
                com[2] += a.x[2]*amass
                mass += amass
        com[0] /= mass
        com[1] /= mass
        com[2] /= mass
        return(com)#,mass)

    def _closest_lig_atom_to_COM( self, system, resname, ligCOM ):
        indLig = -42
        mindist = 99999.999
        # and also search for alternatives if no carbon is found
        mindistAlt = 99999.999
        indLigAlt = -42

        for a in system.atoms:
            if a.resname==resname and (a.id-1 in self.indLig):
                aname = self._get_atom_name( a.name )
                d = np.power(a.x[0]-ligCOM[0],2)+np.power(a.x[1]-ligCOM[1],2)+np.power(a.x[2]-ligCOM[2],2)
                # check for carbon
                if aname.startswith('C') and not aname.startswith('CL'):
                    if d<mindist:
                        mindist = d
                        indLig = a.id-1
                # check for any other atom
                else:
                    if d<mindistAlt:
                        mindistAlt = d
                        indLigAlt = a.id-1
        if indLig==-42:
            indLig = indLigAlt
        return(indLig)

    def _furthest_lig_atom( self, system, resname, indLig1, indLig2=False ):
        indLig = -42
        maxdist = -99999.999
        # and also search for alternatives if no carbon is found
        maxdistAlt = -99999.999
        indLigAlt = -42
        for a in system.atoms:
            if a.resname==resname and (a.id-1 in self.indLig):
                aname = self._get_atom_name( a.name )
                d = a - system.atoms[indLig1]
                if indLig2!=False:
                    d += a - system.atoms[indLig2]
                # check for carbon
                if aname.startswith('C') and not aname.startswith('CL'):
                    if d>maxdist:
                        maxdist = d
                        indLig = a.id-1
                # check for any other atom
                else:
                    if d>maxdistAlt:
                        maxdistAlt = d
                        indLigAlt = a.id-1
        if indLig==-42:
            indLig = indLigAlt
        return(indLig)

    def _closest_lig_atom( self, system, resname, indLig1, usedID ):
        indLig = -42
        mindist = 99999.999
        # and also search for alternatives if no carbon is found
        mindistAlt = 99999.999
        indLigAlt = -42
        for a in system.atoms:
            if a.resname==resname and (a.id-1 in self.indLig):
                aname = self._get_atom_name( a.name )
                d = a - system.atoms[indLig1]
                # check for carbon
                if aname.startswith(('C','N')) and not aname.startswith('CL') and (a.id-1 not in usedID):
                    if d<mindist:
                        mindist = d
                        indLig = a.id-1
                # check for any other atom
                elif a.id-1 not in usedID:
                    if d<mindistAlt:
                        mindistAlt = d
                        indLigAlt = a.id-1
        if indLig==-42:
            indLig = indLigAlt
        return(indLig)

    def _get_resname( self, system ):
        # don't use topology resname, as it may not match the one in the pdb
#        if self.ligTop != None:
#            return(self.ligTop.residues[0].resname)
#        else:
        for a in system.atoms:
            if (a.id-1 in self.indLig):
                return(a.resname)

    def _identify_ligand_atoms( self, system, indLig1 ):
        if indLig1==False:
            indLig1 = -42
        indLig2 = -42
        indLig3 = -42
        usedID = []

        # first atom closest to COM
#        resname = self.ligTop.residues[0].resname
        resname = self._get_resname( system )
        if indLig1==-42:
            ligCOM = self._get_lig_COM( resname, system ) # conversion a2nm is performed
            indLig1 = self._closest_lig_atom_to_COM( system, resname, ligCOM )
        usedID.append(indLig1)            

        if self.bLigMaxSpread==True:
            # second atom furthest from first atom
            indLig2 = self._furthest_lig_atom( system, resname, indLig1 )
        else:
            # second atom closest to the first atom
            indLig2 = self._closest_lig_atom( system, resname, indLig1, usedID )
        usedID.append(indLig2)

        if self.bLigMaxSpread==True:
            # third atom furthest from first and second atoms
            indLig3 = self._furthest_lig_atom( system, resname, indLig1, indLig2 )
        else:
            indLig3 = self._closest_lig_atom( system, resname, indLig1, usedID )
       
        return([indLig1],[indLig2],[indLig3])

    def _ligAtomByName( self, atomName, system ):
        resname = self._get_resname( system )
        #resname = self.ligTop.residues[0].resname
        for a in system.atoms:
            if a.resname==resname and a.name==atomName:
                return( a.id-1 )
        return(False)

    def _dist_between_atoms( self, a1, a2 ):
        d = np.power(a1.x[0]-a2.x[0],2)+np.power(a1.x[1]-a2.x[1],2)+np.power(a1.x[2]-a2.x[2],2)
        return(np.sqrt(d))

    def _closest_prot_atom( self, system, indLig, usedID, bAngleCheck=0, indProt1=-42 ):
        mindist = 999999.999
        indProt = -42
        ligAtom = system.atoms[indLig]
        for a in system.atoms:
            a.a2nm()
            if (a.resname in library._protein_residues) or self.bAnyHeavyAtom==True: # for martini or another ligand it can be anything
                if (a.name=='CA' or a.name=='BB' or self.bAnyHeavyAtom==True) and (a.id-1 not in usedID) and (a.id-1 in self.indProt):
                    d = a - ligAtom 
                    bAngle = True
                    if bAngleCheck==1: # check for angle 1
                        ang = ligAtom.angle(self.aLig2,a,degree=True)
                        bAngle = self._check_angle( ang )
                    elif bAngleCheck==2: # check for angle 2
                        aProt1 = system.atoms[indProt1]
                        ang = aProt1.angle(self.aLig1,a,degree=True)
                        bAngle = self._check_angle( ang )
                    if d<mindist and bAngle:
                        mindist = d
                        indProt = a.id-1
        return(indProt)

    def _identify_protein_atoms( self, system, indLig1, indLig2, indLig3 ):
        indProt1 = -42
        indProt2 = -42
        indProt3 = -42
        indLig1 = indLig1[0]
        indLig2 = indLig2[0]
        indLig3 = indLig3[0]
        usedID = []

        # indProt1 prot atom closest to indLig1
        indProt1 = self._closest_prot_atom( system, indLig1, usedID, bAngleCheck=1 ) 
        usedID.append(indProt1)

        # indProt2 prot atom closest to indLig2
        indProt2 = self._closest_prot_atom( system, indLig2, usedID ) 
        usedID.append(indProt2)

        # indProt3 prot atom closest to indLig3
        indProt3 = self._closest_prot_atom( system, indLig3, usedID, bAngleCheck=2, indProt1=indProt1 ) 
        usedID.append(indProt3)
#        print indLig1, indProt1, indProt2, indProt3

        return([indProt1],[indProt2],[indProt3])

    def _identify_protein_atomsAlt( self, system, indLig ):
        indProt1 = -42
        indProt2 = -42
        indProt3 = -42
        ligAtom = system.atoms[indLig[0]]
        mindist = 99999.999
        minres = -42
        for a in system.atoms:
            if a.resname in library._protein_residues:
                if a.name=='N' or a.name=='CA' or a.name=='C' or a.name=='BB':
                    d = a-ligAtom
                    if d<mindist:
                        mindist = d
                        indProt1 = a.id-1
                        minres = a.resnr-1

        for a in system.residues[minres].atoms:
            if a.name=='N' or a.name=='CA' or a.name=='C' or a.name=='BB':
                if indProt2==-42:
                    if indProt1!=a.id-1:
                        indProt2 = a.id-1
                elif indProt3==-42:
                    if indProt1!=a.id-1:
                        indProt3 = a.id-1

        return([indProt1],[indProt2],[indProt3])


    def _check_ndx_size( self,ndxGroupSize, desiredSize ):
        if ndxGroupSize!=desiredSize:
            sys.stdout.write('Need to select a group with %d atoms, but the selected group contained %d atoms. Exiting...\n' % (desiredSize,ndxGroupSize) )
            sys.exit(0)


    def _select_ndx( self, fname="index.ndx", message=False ):
        if not os.path.isfile( fname ):
            return(False)
        ndx_file = ndx.IndexFile( fname )
        names = ndx_file.names
        i = 0
        ndxDict = {}
        for name in names:
            atomNum = len(ndx_file[name].ids)
            sys.stdout.write('%d %s: %d atoms\n' % (i,name,atomNum) )
            ndxDict[i] = ndx_file[name].ids
            i+=1
        sys.stdout.write('\n')

        if message==False:
            sys.stdout.write('Select a group for analysis:\n')
        else:
            sys.stdout.write(message+'\n')

        ndxNum = -1
        while ndxNum==-1:
            ndxNum = input()
            if ndxNum.isdigit()==False:
                sys.stdout.write('Wrong index group number selected (use an integer number)\n')
                ndxNum = -1
                continue

            ndxNum = int(ndxNum)
            if (ndxNum >= i) or (ndxNum < 0):
                sys.stdout.write('Wrong index group number selected\n')
                ndxNum = -1
        sys.stdout.write('Selected group %d\n\n' % ndxNum)

        res = []
        res = np.asarray(ndxDict[ndxNum])-1 # starting from 0
        return(res)

########################################################################
####### functions to transfer the restraints between two systems #######
########################################################################
def _find_atom_inY( ind, strX, strY ):
    aX = strX.atoms[ind-1]
    aY = 0
    minDist = 999.99
    for a in strY.atoms:
        d = a-aX
        if d<minDist:
            minDist = d
            aY = a
    return(aY.id)

def transfer_ii( pdbX, pdbY, iiInpFname, iiOutFname ):
    strX = Model(pdbX)
    strY = Model(pdbY)

    fp = open(iiInpFname,'r')
    lines = fp.readlines()
    fp.close()

    fp = open(iiOutFname,'w')
    for l in lines:
        if '[' in l:
            fp.write(l)
        elif l=='\n':
            fp.write(l)
        else:
            l = l.rstrip()
            l = l.lstrip()
            foo = l.split()
#             print(foo,len(foo))

            # go over the array in reverse order to get atomID
            out = []
            for i in range(1,len(foo)+1):
                if i>5:
                    atomIndY = _find_atom_inY( int(foo[-i]), strX, strY )
                    out.insert(0,atomIndY)
                else:
                    out.insert(0,foo[-i])

            # go over the elements again to output them
            fp.write('   ')
            for i in range(0,len(foo)):
                fp.write('%s   ' % str(out[i]))
            fp.write('\n')
             
    fp.close()





#!/usr/bin/env python

# pmx  Copyright Notice
# ============================
#
# The pmx source code is copyrighted, but you can freely use and
# copy it as long as you don't change or remove any of the copyright
# notices.
#
# ----------------------------------------------------------------------
# pmx is Copyright (C) 2006-2013 by Daniel Seeliger
#
#                        All Rights Reserved
#
# Permission to use, copy, modify, distribute, and distribute modified
# versions of this software and its documentation for any purpose and
# without fee is hereby granted, provided that the above copyright
# notice appear in all copies and that both the copyright notice and
# this permission notice appear in supporting documentation, and that
# the name of Daniel Seeliger not be used in advertising or publicity
# pertaining to distribution of the software without specific, written
# prior permission.
#
# DANIEL SEELIGER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
# SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS.  IN NO EVENT SHALL DANIEL SEELIGER BE LIABLE FOR ANY
# SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
# CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# ----------------------------------------------------------------------

"""This module contains the Atom class.

Examples
--------

Basic usage:

    >>> model = Model().read(args['-f'])             # read structure file
    >>> atom = model.atoms[0]                        # the first atom
    >>> a = Atom(name='DUM',x = [1,2,3], bfac = 20.) # create an atom

Geometric measurements:

    >>> d = atom1 - atom2                       # calculate distance
    >>> a = atom1.angle(atom2,atom3)            # calculate angle
    >>> dih = atom1.dihedral(atom2,atom3,atom4) # calculate dihedral
    >>> print atom                              # print atom in PDB format

"""

import copy
from numpy import pi
from . import library
from . import _pmx as _p
from .library import pdb_format, pdb_format2

__all__ = ['Atom']


class Atom:
    """Atom Class.

    Parameters
    ----------
    line : str, optional
        input line in PDB format. Default is None.
    mol2line : str, optional
        input line in MOL2 format.  Default is None.


    Attributes
    ----------
    id : int
        atom id
    name : str
        atom name
    resname : str
        name of residue atom is part of
    chain_id : str
        ID of the chain atom belongs to
    x : array
        atom coordinates
    occ : float
        occupancy
    bfac : float
        B-factor
    symbol : str
        chemical element of the atom
    unity : str
        unit of the coordinates, either 'A' or 'nm'. Default is 'A'.
    """
    def __init__(self, line=None, mol2line=None, **kwargs):

        self.race = 'ATOM  '
        self.id = 0
        self.orig_id = 0
        self.code = 0
        self.name = ''
        self.altloc = ''
        self.resname = ''
        self.chain_id = ' '
        self.x = [0, 0, 0]
        self.occ = 1.
        self.bfac = 0.
        self.vdw = 0.
        self.vdw14 = 0.
        self.atype = ''
        self.hyb = ''
        self.symbol = ''
        self.bonds = []
        self.b13 = []
        self.b14 = []
        self.connected = []
        self.neighbors = []
        self.order = 0
        self.boxid = 0
        self.molecule = None
        self.chain = None
        self.model = None
        self.cgnr = 0.
        self.v = [0, 0, 0]
        self.f = []
        self.m = 0.
        self.q = 0.
        self.mB = 0.
        self.qB = 0.
        self.type = ''
        self.typeB = ''
        self.resnr = 0
        self.grpnr = ' '
        self.atomtype = ''
        self.atomtypeB = ''
        self.ptype = ''
        self.long_name = ''
        self.unity = 'A'
        for key, val in kwargs.items():
            setattr(self, key, val)
        if line is not None:
            self.readPDBString(line)
        if mol2line is not None:
            self.read_mol2_line(mol2line)

    def __str__(self):
        """Prints the Atom in PDB format"""
        if self.unity == 'nm':
            coords = list(map(lambda x: x*10, self.x))
        else:
            coords = self.x
        if len(self.resname) < 4:
            resname = self.resname+' '  # [0:3]
        else:
            resname = self.resname
        if len(self.name) == 1:
            name = ' '+self.name+'  '
        elif len(self.name) == 2:
            if self.name[0].isdigit():
                name = self.name+'  '
            else:
                name = ' '+self.name+' '
        elif len(self.name) == 3:
            if self.name[0].isdigit():
                name = self.name+' '
            else:
                name = ' '+self.name
        else:
            name = self.name
        idx = self.id % 100000
        try:
            resid = self.resnr % 10000
        except:
            resid = str(self.resnr)
        try:
            s = pdb_format % (self.race, idx, name, self.altloc,
                              resname, self.chain_id, resid,
                              coords[0], coords[1],
                              coords[2], self.occ, self.bfac, self.symbol)
        except:
            s = pdb_format2 % (self.race, idx, name, self.altloc,
                               resname, self.chain_id, resid,
                               coords[0], coords[1], coords[2],
                               self.occ, self.bfac)
        return s

    def __sub__(self, other):
        """ Overloading of the '-' operator for using
        atom1-atom2 instead of atom1.dist(atom2)"""
        return self.dist(other)

    def readPDBString(self, line, origID=0):
        """PDB String to Atom"""

        self.race = line[0:6]
        self.id = int(line[7:11])
        self.orig_id = origID
        self.name = line[12:16].strip()
        self.altloc = line[16]
        self.resname = line[17:21].strip()
        self.chain_id = line[21]
        try:
            self.resnr = int(line[22:27])
        except:
            self.resnr = line[22:27].lstrip().rstrip()  # contains insertion code

        self.x = [float(line[30:38]),
                  float(line[39:46]),
                  float(line[47:54])]
        try:
            self.occ = float(line[55:60])
        except:
            self.occ = 1.
        try:
            self.bfac = float(line[61:66])
        except:
            self.bfac = 0.
        try:
            self.symbol = line[70:73].strip()
        except:
            self.symbol = None
        self.unity = 'A'
        if not self.symbol:
            self.get_symbol()
        return self

    def read_mol2_line(self, line):
        entr = line.split()
        if len(entr) == 9:
            self.id = int(entr[0])
            self.name = entr[1]
            self.x[0] = float(entr[2])
            self.x[1] = float(entr[3])
            self.x[2] = float(entr[4])
            self.atype = entr[5]
            self.resnr = int(entr[6])
            self.resname = entr[7]
            self.q = float(entr[8])
            self.unity = 'A'
            self.symbol = self.atype.split('.')[0]
        elif len(entr) == 10:
            self.id = int(entr[0])
            self.name = entr[1]
            self.x[0] = float(entr[3])
            self.x[1] = float(entr[4])
            self.x[2] = float(entr[5])
            self.atype = entr[6]
            self.resnr = int(entr[7])
            self.resname = entr[8]
            self.q = float(entr[9])
            self.unity = 'A'
            self.symbol = self.atype.split('.')[0]
        else:
            print(line)
            raise ValueError('Error: Cannot convert line to atom')
        return self

    def make_long_name(self):
        """Make extended name to determine element and order.
        """
        # check for aliases first
        ali = library._aliases
        if self.resname in ali and self.name in ali[self.resname]:
            name = ali[self.resname][self.name]
        else:
            name = self.name.strip()
        if len(name) == 4:
            if not name[0].isdigit():
                tmp = name[3]
                name = tmp+name[:-1]
        elif len(name) == 3:
            if name[0].isdigit():
                name += ' '
            else:
                name = ' '+name
        elif len(name) == 2:
            if name[1].isdigit() and name[0] == 'H' and \
                   self.resname in library._aacids:
                name = name[1]+name[0]+'  '
            else:
                name = ' ' + name + ' '
        elif len(name) == 1:
            name = ' '+name+'  '
        self.long_name = name

    def copy(self):
        """copy atom"""
        return copy.deepcopy(self)

    def get_symbol(self):
        """Set atom element"""
        if self.long_name == '':
            self.make_long_name()
        if self.resname in library._protein_residues or \
           self.resname in library._nucleic_acids:
            self.symbol = self.long_name[1]
        elif self.resname in library._ions:
            self.symbol = self.resname.upper()
            if len(self.symbol) > 2:
                self.symbol = self.symbol[:2]
        elif self.resname in library._water:
            self.symbol = self.name[0]
        else:
            # unknown residue
            c1 = self.long_name[1]
            c2 = self.long_name[2]
            if c1 == 'C':
                self.symbol = 'C'
                if c2.upper() == 'L':
                    self.symbol = 'CL'
            elif c1 == 'O':
                self.symbol = 'O'
            elif c1 == 'N':
                self.symbol = 'N'
            elif c1 == 'H':
                self.symbol = 'H'
            elif c1 == 'S':
                self.symbol = 'S'
            elif c1 == 'P':
                self.symbol = 'P'
            elif c1 == 'B':
                if c2.upper() == 'R':
                    self.symbol = 'BR'
            elif c1 == 'F':
                if c2.upper() == 'E':
                    self.symbol = 'FE'
                else:
                    self.symbol = 'F'
            elif c1 == 'I':
                self.symbol = 'I'
            elif c1 == 'D':
                self.symbol = 'D'
            elif c1 == 'M':
                self.symbol = 'D'
            else:
                self.symbol = 'UN'

    def get_order(self):
        """Get the order (number of bonds to mainchain)"""
        if self.long_name == '':
            self.make_long_name()
        if self.symbol == '':
            self.get_symbol()
        if self.resname not in library._protein_residues:
            print('Sorry, implemented for proteins only')
            return

        el = self.symbol
        x = self.long_name[2]
        if self.name in ['C', 'CA', 'N']:
            self.order = 0
        elif self.name in ['H', 'H1', 'H2', 'H3', '1H', '2H',
                           '3H', 'O', 'O1',
                           'O2', 'OC1', 'OC2', 'OXT',
                           'OT1', 'OT2']:
            self.order = 1
        else:
            if el != 'H':
                if x == 'B':
                    self.order = 1
                elif x == 'G':
                    self.order = 2
                elif x == 'D':
                    self.order = 3
                elif x == 'E':
                    self.order = 4
                elif x == 'Z':
                    self.order = 5
                elif x == 'H':
                    self.order = 6
            elif el == 'H':
                if x == 'A':
                    self.order = 1
                elif x == 'B':
                    self.order = 2
                elif x == 'G':
                    self.order = 3
                elif x == 'D':
                    self.order = 4
                elif x == 'E':
                    self.order = 5
                elif x == 'Z':
                    self.order = 6
                elif x == 'H':
                    self.order = 7

    # ==============
    # Public Methods
    # ==============
    def dist(self, other):
        """Calculates the distance between two atoms. This function is also when
        subtracting two atom instances, as in the example.

        Examples
        --------
        >>> dist = atom1.dist(atom2)
        >>> dist = atom1 - atom2

        Returns
        -------
        dist : float
            distance between two Atom instances.
        """
        return _p.dist(self.x, other.x)

    def dist2(self, other):
        """Calculates the squared distance between two atoms.

        Examples
        --------
        >>> dist = atom1.dist2(atom2)

        Returns
        -------
        dist2 : float
            squared of the distance between two Atom instances
        """
        return _p.dist2(self.x, other.x)

    def translate(self, v):
        """Translates the position of the atom.

        Parameters
        ----------
        v : array
            vector of length 3 containing the values by which to translate the
            atom in the x, y, and z dimensions.
        """
        self.x[0] += v[0]
        self.x[1] += v[1]
        self.x[2] += v[2]

    def nm2a(self):
        """Converts the Atom coordinates from nanometers (nm) to Angstrom (A).
        """
        if self.unity == 'nm':
            self.x[0] *= 10
            self.x[1] *= 10
            self.x[2] *= 10
            self.unity = 'A'

    def a2nm(self):
        """Converts the Atom coordinates from Angstrom (A) to nanometers (nm).
        """
        if self.unity == 'A':
            self.x[0] *= .1
            self.x[1] *= .1
            self.x[2] *= .1
            self.unity = 'nm'

    def angle(self, other1, other2, degree=False):
        """Calcluates the angle between 3 atoms. Returns radians by default.

        Parameters
        ----------
        other1 : Atom
            second atom
        other2 : Atom
            third atom
        degree : bool
            whether to return the angle in degrees rather than radians. Default
            is False.

        Examples
        --------
        >>> atom1.angle(atom2, atom3)
        >>> 1.57079632679
        >>> atom1.angle(atom2, atom3, degree=True)
        >>> 90.0

        Notes
        -----
        Note that the angle being calculated is the one where the Atom instance
        is in the middle. Following the example above: *atom2 - atom1 - atom3*.

        Returns
        -------
        angle : float
            the value of the angle
        """

        # Note: atom1 must be between 2 and 3
        angle = _p.angle(other1.x, self.x, other2.x)
        if degree is True:
            return angle*180.0/pi
        else:
            return angle

    def dihedral(self, other1, other2, other3, degree=False):
        """Calculates the dihedral angle between four atoms. Returns radians
        by default.

        Parameters
        ----------
        other1 : Atom
            second atom
        other2 : Atom
            third atom
        other3 : Atom
            fourth atom
        degree : bool
            whether to return the angle in degrees rather than radians. Default
            is False.

        Examples
        --------
        >>> atom1.dihedral(atom2, atom3, atom4)
        >>> 1.57079632679
        >>> atom1.dihedral(atom2, atom3, atom4, degree=True)
        >>> 90.0

        Returns
        -------
        angle : float
            the value of the dihedral angle
        """
        ang = _p.dihedral(self.x, other1.x, other2.x, other3.x)
        if degree is True:
            return ang*180.0/pi
        else:
            return ang

    def set_resname(self, resname):
        """Change the residue name.
        """
        self.resname = resname

    def set_chain_id(self, chain_id):
        """Change the chain identifier.
        """
        self.chain_id = chain_id



#!/usr/bin/env python
# pmx  Copyright Notice
# ============================
#
# The pmx source code is copyrighted, but you can freely use and
# copy it as long as you don't change or remove any of the copyright
# notices.
#
# ----------------------------------------------------------------------
# pmx is Copyright (C) 2006-2013 by Daniel Seeliger
#
#                        All Rights Reserved
#
# Permission to use, copy, modify, distribute, and distribute modified
# versions of this software and its documentation for any purpose and
# without fee is hereby granted, provided that the above copyright
# notice appear in all copies and that both the copyright notice and
# this permission notice appear in supporting documentation, and that
# the name of Daniel Seeliger not be used in advertising or publicity
# pertaining to distribution of the software without specific, written
# prior permission.
#
# DANIEL SEELIGER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
# SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS.  IN NO EVENT SHALL DANIEL SEELIGER BE LIABLE FOR ANY
# SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
# CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# ----------------------------------------------------------------------

import sys,os
from pmx import *
from pmx.options import *
from pmx.parser import *
from pmx import library
from pmx.mutdb import *
from pmx.geometry import *

def rename_lowlevel(r,name):
    for a in r.atoms:
         a.resname = name

def rename_HA( a ):
    if (a.name=='HA2'):
        a.name = 'HA1'
    if (a.name=='HA3'):
        a.name = 'HA2'

def rename_HB( a ):
    if (a.name=='HB2'):
        a.name = 'HB1'
    if (a.name=='HB3'):
        a.name = 'HB2'

def rename_HG( a ):
    if (a.name=='HG2'):
        a.name = 'HG1'
    if (a.name=='HG3'):
        a.name = 'HG2'
    if (a.name=='HG12'):
        a.name = 'HG11'
    if (a.name=='HG13'):
        a.name = 'HG12'
    if (a.name=='2HG1'):
        a.name = '1HG1'
    if (a.name=='3HG1'):
        a.name = '2HG1'

def rename_HD( a ):
    if (a.name=='HD2'):
        a.name = 'HD1'
    if (a.name=='HD3'):
        a.name = 'HD2'
    if a.resname=='ILE':
        if a.name=='1HD1':
            a.name = 'HD1'
        if a.name=='2HD1':
            a.name = 'HD2'
        if a.name=='3HD1':
            a.name = 'HD3'

def rename_HE( a ):
    if (a.name=='HE2') and ('HI' not in a.resname):
        a.name = 'HE1'
    if (a.name=='HE3'):
        a.name = 'HE2'

def rename_HH( a ):
    if (a.name=='2HH1'):
        a.name = 'HH12'
    if (a.name=='1HH1'):
        a.name = 'HH11'
    if (a.name=='1HH2'):
        a.name = 'HH21'
    if (a.name=='2HH2'):
        a.name = 'HH22'

def rename_HE_TRP( a ):
    if (a.name=='HE2'):
        a.name = 'HE3'

def rename_CT3( a ):
    if (a.name=='CA'):
        a.name = 'CH3'
    if (a.name=='HA1') or (a.name=='1HA'):
        a.name = 'HH31'
    if (a.name=='HA2') or (a.name=='2HA'):
        a.name = 'HH32'
    if (a.name=='HA3') or (a.name=='3HA'):
        a.name = 'HH33'


def rename_hydrogens(m,ff,bMaestro=False):
    for a in m.atoms:
	# Maestro
        if bMaestro:
            if( a.resname=='ALA'):
                continue
            elif( a.resname=='ARG'):
                rename_HB( a )
                rename_HG( a )
                rename_HD( a )
                rename_HH( a )
            elif( a.resname=='ASN'):
                rename_HB( a )
            elif( a.resname=='ASP'):
                rename_HB( a )
            elif( a.resname=='ASPP'):
                rename_HB( a )
            elif( a.resname=='ASH'):
                rename_HB( a )
            elif( 'CY' in a.resname):
                rename_HB( a )
            elif( a.resname=='GLY'):
                rename_HA( a )
            elif( a.resname=='GLU'):
                rename_HB( a )
                rename_HG( a )
            elif( a.resname=='GLN'):
                rename_HB( a )
                rename_HG( a )
            elif( 'HI' in a.resname or 'HS' in a.resname ):
                rename_HB( a )
            elif( a.resname=='ILE'):
                rename_HG( a )
                rename_HD( a )
            elif( a.resname=='LEU'):
                rename_HB( a )
            elif( a.resname=='LYS'):
                rename_HB( a )
                rename_HG( a )
                rename_HD( a )
                rename_HE( a )
            elif( a.resname=='LYN' or a.resname=='LSN'):
                rename_HB( a )
                rename_HG( a )
                rename_HD( a )
                rename_HE( a )
            elif( a.resname=='MET'):
                rename_HB( a )
                rename_HG( a )
            elif( a.resname=='PHE'):
                rename_HB( a )
            elif( a.resname=='PRO'):
                rename_HB( a )
                rename_HG( a )
                rename_HD( a )
            elif( a.resname=='SER'):
                rename_HB( a )
#            elif( a.resname=='THR'):
#                continue
            elif( a.resname=='TRP'):
                rename_HB( a )
            elif( a.resname=='TYR'):
                rename_HB( a )
            elif( a.resname=='CT3'):
                rename_CT3( a )
#            elif( a.resname=='VAL'):
#                continue
#            elif( a.resname=='ACE'):
#            elif( a.resname=='NME'):
            if (a.resname=='HOH') and (a.name=='H1'):
                a.name = 'HW1'
            if (a.resname=='HOH') and (a.name=='H2'):
                a.name = 'HW2'
	# H and HN
        if (a.name=='H2' or a.name=='H11' or a.name=='H1') and bMaestro:
            if ff.startswith('charmm'):
                a.name = 'HN'
            else:
                a.name = 'H'
	# H and HN
        if (a.name=='H') and ff.startswith('charmm'):
            a.name = 'HN'
        if (a.name=='H1' or a.name=='1H') and ff.startswith('charmm') and a.resname=='PRO': # terminal proline in charmm
            a.name = 'HN1'
        if (a.name=='H2' or a.name=='2H') and ff.startswith('charmm') and a.resname=='PRO': # terminal proline in charmm
            a.name = 'HN2'
        if (a.name=='HN') and (ff.startswith('opls') or ff.startswith('amber')):
            a.name = 'H'
	# SER
        if (a.resname=='SER') and (a.name=='HG') and ff.startswith('charmm'):
            a.name = 'HG1'
        if (a.resname=='SER') and (a.name=='HG1') and (ff.startswith('opls') or ff.startswith('amber')):
            a.name = 'HG'
	# CYS
        if (a.resname=='CYS') and (a.name=='HG') and ff.startswith('charmm'):
            a.name = 'HG1'
        if (a.resname=='CYS') and (a.name=='HG1') and (ff.startswith('opls') or ff.startswith('amber')):
            a.name = 'HG'
	# ACE
        if (a.resname=='ACE') and (a.name=='H1' or a.name=='1H' or a.name=='H'):
            a.name = 'HH31'
        if (a.resname=='ACE') and (a.name=='H2' or a.name=='2H' or a.name=='H'):
            a.name = 'HH32'
        if (a.resname=='ACE') and (a.name=='H3' or a.name=='3H' or a.name=='H'):
            a.name = 'HH33'
	# NME
        if (a.resname=='NME') and (a.name=='CA' or a.name=='C'):
            a.name = 'CH3'
        if (a.resname=='NME') and (a.name=='HA1' or a.name=='1HA'):
            a.name = 'HH31'
        if (a.resname=='NME') and (a.name=='HA2' or a.name=='2HA'):
            a.name = 'HH32'
        if (a.resname=='NME') and (a.name=='HA3' or a.name=='3HA'):
            a.name = 'HH33'
    


def main(argv):

   options=[
   Option( "-ft", "string", "charmm" , "force field type (charmm, amber99sb, amber99sb*-ildn, oplsaa"),
   Option( "-maestro", "bool", False , "conversion from Maestro prepared structure"),
#   Option( "-align", "bool", True, "align side chains"),
        ]
    
   files = [
       FileOption("-f", "r",["pdb","gro"],"protein.pdb", "input structure file"),
       FileOption("-o", "w",["pdb","gro"],"out.pdb", "output structure file"),
       ]
    
   help_text = ('Rename hydrogens (and some other atoms) to match the force fields, ',
                '',
                )

    
   cmdl = Commandline( argv, options = options,
                       fileoptions = files,
                       program_desc = help_text,
                       check_for_existing_files = False )
    
   iname = cmdl['-f']
   oname = cmdl['-o']
   ff = cmdl['-ft']
   bMaestro = False
   if cmdl['-maestro']==True:
       bMaestro = True

   m = Model(iname,bPDBTER=True,renumber_residues=False)

   rename_hydrogens(m,ff,bMaestro)

   m.write(oname,bPDBTER=True)

if __name__=='__main__':
    main(sys.argv)




#!/usr/bin/env python

# pmx  Copyright Notice
# ============================
#
# The pmx source code is copyrighted, but you can freely use and
# copy it as long as you don't change or remove any of the copyright
# notices.
#
# ----------------------------------------------------------------------
# pmx is Copyright (C) 2006-2013 by Daniel Seeliger
#
#                        All Rights Reserved
#
# Permission to use, copy, modify, distribute, and distribute modified
# versions of this software and its documentation for any purpose and
# without fee is hereby granted, provided that the above copyright
# notice appear in all copies and that both the copyright notice and
# this permission notice appear in supporting documentation, and that
# the name of Daniel Seeliger not be used in advertising or publicity
# pertaining to distribution of the software without specific, written
# prior permission.
#
# DANIEL SEELIGER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
# SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS.  IN NO EVENT SHALL DANIEL SEELIGER BE LIABLE FOR ANY
# SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
# CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# ----------------------------------------------------------------------

"""This module contains the Model class. It can use
GROMACS routines to read and write structure files. Moreover it
allows to modify structure files in various ways. E.g.:

1. Rename atoms, residues, chains
2. Delete or add atoms, residues and chains

By default, all residues are renumbered from 1. This option can be deactivated
by setting renumber_residues to False.

The Model instance contains:

* model.atoms       -> list of atoms
* model.residues    -> list of residues
* model.chains      -> list of chains
* model.chdic       -> chain dictionary (chdic['A'] returns chain A)

Examples
--------
Basic usage:

    >>> model = Model('input.pdb')

Some useful methods:

    >>> # returns all backbone atoms
    >>> model.fetch_atoms(['CA','N','C'])
    >>> # returns all carbon atoms
    >>> model.fetch_atoms('C',how='byelem')
    >>> # return all atoms except hydrogens
    >>> model.fetch_atoms('H',how='byelem',inv=True)

    >>> # return all ALA,TRP and CYS residues
    >>> model.fetch_residues(['ALA','TRP','CYS'])

    >>> # returns the first 10 residues
    >>> rl = model.residues[:10]
    >>> # return the last residue from chain A
    >>> rl = model.chdic['A'].residues[-1]
    >>> # returns a list with the first residues of each chain
    >>> rl = map(lamda m: m.residues[0], model.chains)
    >>> # remove chain A
    >>> del model['A']
    >>> # write new structure file
    >>> model.write('output.pdb')

"""

import sys
import copy
import numpy as np
from . import _pmx as _p
from . import library
from . import chain
from .atomselection import Atomselection
from .molecule import Molecule
from .atom import Atom
from string import digits


__all__ = ['Model']


class Model(Atomselection):
    """Model Class.

    Parameters
    ----------
    filename : str
        filename of input structure
    pdbline : bool(?)
        what is pdbline?
    renumber_atoms : bool, optional
        renumber all atoms from 1. Default is True.
    renumber_residues : bool, optional
        renumber all residues from 1. In this way, each residue will have a
        unique ID, also across chains. Default is True.
    rename_atoms : bool, optional
        rename atoms so to conform to Gromacs format. Default is False.
    scale_coords : A|nm, optional
        whether to enforce the units of the coordinates to be in A or in nm.
        By default, PDB coordinates are assumed to be in Angstrom (A) and
        GRO coordinates in nanometers (nm). If you read a PDB file but would
        like operate on nm coordinates, then select "nm". Viceversa, if you
        read a GRO file but would like to work on A coordinates, select "A".
        Note that if you read a PDB file in A coordinates and select "A",
        nothing happens (same for GRO and "nm" selection).
    bPDBTER : bool
        whether to recognize TER lines and other chain breaks, e.g.
        discontinuous residue indices(?). Default is True.
    bNoNewID : bool
        whether to assign new chain IDs? If True, new chain IDs starting
        with 'pmx' will be assigned(?). Only relevant if bPDBTER is True.
        Default is True.
    bPDBGAP : bool
        whether search for gaps in the chain to assign new chain IDs.
    bPDBMASS : bool
        whether to guess masses from the atom library (will fail for
        complex atom naming, e.g. ND will not be interpreted as nitrogen)

    Attributes
    ----------
    title : str
        title of model. Default is 'PMX MODEL'.
    filename : str
        filename from which the Model was imported, otherwise None.
    chains : list
        list of Chain instances
    chdic : dict
        dict with chain IDs as keys and Chain instances as values
    residues : list
        list of molecules/residues
    unity : str
        coordinates unit, either 'A' or 'nm'.
    box : 2d array
        3x3 array containing the box vectors. See Gromacs manual, Table 3.1
    moltype : str
        Type of system: protein, dna, rna, or unknown if organic molecule or
        a mix of molecules are in the system.
    """
    def __init__(self, filename=None, pdbline=None, renumber_atoms=True,
                 renumber_residues=True, rename_atoms=False, scale_coords=None,
                 bPDBTER=True, bNoNewID=True, bPDBGAP=False, bPDBMASS=False,
                 **kwargs):

        Atomselection.__init__(self)
        self.title = 'PMX MODEL'
        self.filename = filename
        self.chains = []
        self.chdic = {}
        self.residues = []
        self.name = None  # FIXME/QUESTION: self.name not used anywhere? remove?
        self.id = 0
        self.have_bonds = 0  # FIXME/QUESTION: same as above: never used -> remove?
        self.box = [[0, 0, 0], [0, 0, 0], [0, 0, 0]]
        self.unity = 'A'
        for key, val in kwargs.items():
            setattr(self, key, val)

        if filename is not None:
            self.read(filename=filename, bPDBTER=bPDBTER, bNoNewID=bNoNewID, bPDBGAP=bPDBGAP, bPDBMASS=bPDBMASS)
        if pdbline is not None:
            self.__readPDB(pdbline=pdbline)
        if self.atoms:
            self.unity = self.atoms[0].unity
            self.make_chains()
            self.make_residues()
        if self.residues and not self.atoms:
            self.al_from_resl()
            self.make_chains()
            self.make_residues()
        if self.chains and not self.residues:
            self.resl_from_chains()
            self.al_from_resl()
            self.make_chains()
            self.make_residues()
        if self.chdic and not self.chains:
            for key, val in self.chdic.items():
                self.chains.append(val)
            if not self.atoms and not self.residues:
                self.resl_from_chains()
                self.al_from_resl()
        if renumber_atoms is True:
            self.renumber_atoms()
        if renumber_residues is True:
            self.renumber_residues()
        if rename_atoms is True:
            self.rename_atoms_to_gmx()
        if scale_coords is not None:
            if scale_coords == 'A':
                self.nm2a()
            elif scale_coords == 'nm':
                self.a2nm()
            else:
                raise ValueError('unknown unit %s for coordinates' % scale_coords)

        self.assign_moltype()



    def __str__(self):
        s = '< Model: moltype=%s, nchain=%d, nres=%d, natom=%d >' %\
            (self.moltype, len(self.chains), len(self.residues),
             len(self.atoms))
        return s

    def writePIR(self, filename, title=""):
        """Prints sequence to screen in PIR format"""
        fp = open(filename, "w")
        if not title:
            title = '_'.join(self.title.split())
        print('>P1;%s' % title, file=fp)
        print('sequence:::::::::', file=fp)
        for i in range(len(self.chains) - 1):
            print('%s/' % self.chains[i].get_sequence(), file=fp)
        print('%s*' % self.chains[-1].get_sequence(), file=fp)
        fp.close()

    def writeFASTA(self, filename, title=""):
        """Prints sequence to screen in FASTA format"""
        fp = open(filename, "w")
        if not title:
            title = '_'.join(self.title.split())
        if len(self.chains) == 1:
            print('> %s' % title, file=fp)
            print(self.chains[0].get_sequence(), file=fp)
        else:
            for chain in self.chains:
                print('> %s_chain_%s' % (title, chain.id), file=fp)
                print(chain.get_sequence(), file=fp)

    # FIXME/TODO: this function is overwriting the write function inherited from
    # Atomselection. Should we keep only one of the 2?
    # writePIR and writeFASTA could be moved to Atomselection so that all
    # write functions are in one place
    def write(self, fn, title='', nr=1, bPDBTER=False, bAssignChainIDs=False):
        """Writes Model to file. The format is deduced from the filename
        extension given, and available options are '.gro', '.pdb', '.fasta',
        '.pir'.

        Parameters
        ----------
        fn : str
            filename
        title : str, optional
            title of the Model to write in fn
        nr : int, optional
            what is nr?
        bPDBTER : bool, optional
            whether to separate chains with TER entries. Only relevant when
            writing PDB files. Default is False.
        bAssignChainIDs : bool, optional
            whether to write the chains IDs to file. Only relevant when
            writing PDB files. Default is False.
        """
        ext = fn.split('.')[-1]
        if ext == 'pdb':
            self.writePDB(fn, title, nr, bPDBTER, bAssignChainIDs)
        elif ext == 'gro':
            self.writeGRO(fn, title)
        elif ext == 'pir':
            self.writePIR(fn, title)
        elif ext == 'fasta':
            self.writeFASTA(fn, title)
        else:
            print('pmx_Error_> Can only write pdb or gro!', file=sys.stderr)
            sys.exit(1)

    def make_chains(self):
        """Initialises the Chain instances from the input"""
        self.chains = []
        self.chdic = {}
        cur_chain = None
        ch = None
        for atom in self.atoms:
            if atom.chain_id == cur_chain:
                if ch:
                    atom.chain = ch
                    ch.atoms.append(atom)
                else:
                    ch = chain.Chain()
                    cur_chain = atom.chain_id
                    ch.id = atom.chain_id
                    atom.chain = ch
                    ch.atoms.append(atom)
            else:
                if ch:
                    self.chains.append(ch)
                    ch = chain.Chain()
                    cur_chain = atom.chain_id
                    ch.id = cur_chain
                    atom.chain = ch
                    ch.atoms.append(atom)
                else:
                    ch = chain.Chain()
                    cur_chain = atom.chain_id
                    ch.id = atom.chain_id
                    atom.chain = ch
                    ch.atoms.append(atom)

        self.chains.append(ch)
        for ch in self.chains:
            ch.model = self
            idx = ch.id
            self.chdic[idx] = ch

    def make_residues(self):
        """Initialises the Molecule instances from the input"""
        self.residues = []
        for ch in self.chains:
            cur_mol = None
            mol = None
            for atom in ch.atoms:
                if atom.resnr == cur_mol:
                    if mol:
                        mol.atoms.append(atom)
                    else:
                        mol = Molecule()
                        cur_mol = atom.resnr
                        mol.resname = atom.resname
                        mol.id = cur_mol
                        mol.atoms.append(atom)
                else:
                    if mol:
                        mol.model = self
                        mol.chain = ch
                        ch.residues.append(mol)
                        self.residues.append(mol)
                        mol = Molecule()
                        cur_mol = atom.resnr
                        mol.resname = atom.resname
                        mol.id = cur_mol
                        mol.atoms.append(atom)
                    else:
                        mol = Molecule()
                        cur_mol = atom.resnr
                        mol.resname = atom.resname
                        mol.id = cur_mol
                        mol.atoms.append(atom)
            self.residues.append(mol)
            ch.residues.append(mol)
        for r in self.residues:
            r.assign_moltype()
            for atom in r.atoms:
                atom.molecule = r
                atom.model = self
                r.model = self
        for ch in self.chains:
            for r in ch.residues:
                r.chain = ch
                r.chain_id = ch.id

    def __readPDB(self, fname=None, pdbline=None):
        """Reads a PDB file"""
        if pdbline:
            lines = pdbline.split('\n')
        else:
            lines = open(fname, 'r').readlines()
        for line in lines:
            if line[:4] == 'ATOM' or line[:6] == 'HETATM':
                a = Atom().readPDBString(line)
                self.atoms.append(a)
            if line[:6] == 'CRYST1':
                self.box = _p.box_from_cryst1(line)
        self.make_chains()
        self.make_residues()
        self.unity = 'A'
        return self

    def __check_if_gap( self, atC, atN ):
        if atC==None:
            return(False)
        if atN.name != 'N':
            return(False)
        d = atC - atN
        if d > 1.7: # bond 
            return(True)
        return(False)

    def __compareWithoutLastChar(self, str1, str2): 
        if isinstance(str1,int) and isinstance(str2,int): # e.g. 52, 53
            return(False)
 
        if isinstance(str1,int): # e.g. 52, 52A
            if str1==int(str2[0:-1]): # 52, 52A
                return(True)
            if str1==int(str2[0:-1])-1: # 52, 53A
                return(True)
        elif isinstance(str2,int): # e.g. 52A, 52
            if int(str1[0:-1])==str2: # 52A, 52
                return(True)
            if int(str1[0:-1])==str2-1: # 52A, 53
                return(True)
        else: # e.g. 52A, 52B
            if int(str1[0:-1])==int(str2[0:-1]): # 52A, 52B
                return(True)
            if int(str1[0:-1])==int(str2[0:-1])-1: # 52A, 53B
                return(True)
        return(False)

    def _getChainIDs( self, lines ):
        """from the pdb lines extract chain IDs """
        usedChainIDs = []
        for line in lines:
            if (line[:4] == 'ATOM') or (line[:6] == 'HETATM'):
                a = Atom().readPDBString(line)
                if( a.chain_id not in usedChainIDs ):
                    usedChainIDs.append( a.chain_id )
        return( usedChainIDs )
        
    # TODO: make readPDB and readPDBTER a single function. It seems like
    # readPDBTER is more general PDB reader?
    def __readPDBTER(self, fname=None, pdbline=None, bNoNewID=True, bPDBGAP=False, bPDBMASS=False):
        """Reads a PDB file with more options than __readPDB ?"""
        if pdbline:
            lines = pdbline.split('\n')
        else:
            lines = open(fname, 'r').readlines()

        chainIDstringInit = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnoprstuvwxyz123456789'
        chainIDstring = copy.deepcopy(chainIDstringInit)
        bNewChain = True
        chainID = ' '
        prevID = ' '
        prevAtomName = ' '
        prevResID = 0
        prevResName = ' '
        usedChainIDs = []
        atomcount = 1
        prevCatom = None

        # get already used chain IDs
        usedChainIDs = self._getChainIDs( lines )

        for line in lines:
            if 'TER' in line:
                bNewChain = True
            if (line[:4] == 'ATOM') or (line[:6] == 'HETATM'):
                a = Atom().readPDBString(line, origID=atomcount)
                atomcount += 1
                # identify chain change by ID (when no TER is there)
                if (a.chain_id != prevID):
                    bNewChain = True
                if (self.__check_if_gap( prevCatom,a )==True and bPDBGAP==True):
                    bNewChain = True
                if a.resname=='ACE' and prevResName!='ACE':
                    bNewChain = True
                if (a.resnr != prevResID):
                    try:
                        if self.__compareWithoutLastChar(prevResID,a.resnr)==True: # there are some special cases where residues are named, e.g. 52, 52A, 52B, ...
                            bNewChain = False
                        elif a.resnr != prevResID+1:
                            bNewChain = True
                        if (prevAtomName == 'OC2') or (prevAtomName == 'OXT') or (prevAtomName == 'OT2'):
                            bNewChain = True
                        if (prevAtomName == 'HH33') and ((prevResName=='NME') or (prevResName=='NAC') or (prevResName=='CT3')): # NME cap
                            bNewChain = True
                        # do not assign new chain IDs for waters, ions 
                        if (a.resname=='WAT') or (a.resname=='SOL') or (a.resname=='TIP3') or (a.resname=='HOH') \
                           or (a.resname=='NA') or (a.resname=='CL') \
                           or (a.resname=='NaJ') or (a.resname=='Na') or (a.resname=='Cl') or (a.resname=='K') or (a.resname=='KJ') \
                           or (a.resname=='MG') or (a.resname=='Mg') or (a.resname=='CA') or (a.resname=='Ca') or (a.resname=='CaJ') \
                           or (a.resname=='ZN') or (a.resname=='Zn'): # add other ions when needed
                            bNewChain = False
                            chainID = ''
                    except TypeError:
                        bNewChain = False
                        chainID = ''
                # the option below overwrites all the others:
                # when a non-empty ID of the next chain is the same as the previous, then do no change
                if (a.chain_id!='') and (a.chain_id!=' ') and (a.chain_id==prevID):
                    bNewChain = False
                prevID = a.chain_id
                prevResID = a.resnr
                prevAtomName = a.name
                prevResName = a.resname
                if a.name == 'C':
                    prevCatom = a
                if bNewChain==True:
                    if ((a.chain_id==' ') or (a.chain_id==chainID) or (a.chain_id in usedChainIDs) and bNoNewID==False):
                        # find a new chain id
                        bFound = False
                        while bFound==False:
                            if len(chainIDstring)==0: # used up all the IDs
                                chainIDstring = copy.deepcopy(chainIDstringInit)
                                chainID = "pmxX"
                                break
                            foo = chainIDstring[0]
                            chainIDstring = chainIDstring.lstrip(chainIDstring[0])
                            if foo not in usedChainIDs:
                                bFound=True
                                chainID = foo
                                if bNoNewID==True:
                                    chainID = "pmx"+foo
                                usedChainIDs.append(chainID)
                    else:
                        chainID = a.chain_id
                        usedChainIDs.append(chainID)
                a.chain_id = chainID
                self.atoms.append(a)
                bNewChain = False
            if line[:6] == 'CRYST1':
                self.box = _p.box_from_cryst1(line)

        ##### now fix chain IDs that have been newly created #####
        newChainDict = {}
        chainIDstring = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnoprstuvwxyz123456789'
        for a in self.atoms:
            # chain with a new ID
            if 'pmx' in a.chain_id:
                # this ID has already been encountered
                if a.chain_id in newChainDict.keys():
                    a.chain_id = newChainDict[a.chain_id]
                # ID not yet encountered
                else:
                    # find a suitable ID
                    bFound = False
                    while bFound==False:
                        if len(chainIDstring)==0: # used up all the IDs
                            chainIDstring = copy.deepcopy(chainIDstringInit)
                            newChainDict[a.chain_id] = "X"
                            a.chain_id = "X"
                            break
                        foo = chainIDstring[0]
                        chainIDstring = chainIDstring.lstrip(chainIDstring[0])
                        # found
                        if foo not in usedChainIDs:
                            bFound=True
                            usedChainIDs.append(foo)
                            newChainDict[a.chain_id] = foo
                            a.chain_id = foo

        self.make_chains()
        self.make_residues()
        self.unity = 'A'

        if bPDBMASS==True:
            assign_masses_to_model( self )

        return self

    def __readGRO(self, filename):
        """Reads a GRO file"""
        l = open(filename).readlines()
        # first line is name/comment
        name = l[0].rstrip()
        self.title = name
        # next line is number of atoms
        natoms = int(l[1])
        atoms_parsed = 0
        while atoms_parsed != natoms:
            line = l[atoms_parsed+2]
            resid = int(line[:5])
            resname = line[5:9].strip()
            name = line[10:15].strip()
            idx = int(line[15:20])
            rest = line[20:].split()
            assert len(rest) in [3, 6]
            x = float(rest[0])
            y = float(rest[1])
            z = float(rest[2])
            coords = [x, y, z]
            if len(rest) == 6:
                vx = float(rest[3])
                vy = float(rest[4])
                vz = float(rest[5])
                vel = [vx, vy, vz]
            else:
                vel = [0, 0, 0]
            a = Atom(id=idx, name=name, resname=resname,
                     resnr=resid, x=coords, v=vel, unity='nm')
            a.get_symbol()
            self.atoms.append(a)
            atoms_parsed += 1
        box_line = [float(i) for i in l[-1].split()]
        assert len(box_line) in [3, 9]
        box = [[0, 0, 0], [0, 0, 0], [0, 0, 0]]
        box[0][0] = box_line[0]
        box[1][1] = box_line[1]
        box[2][2] = box_line[2]
        if len(box_line) == 3:
            box[0][1] = 0
            box[0][2] = 0
            box[1][0] = 0
            box[1][2] = 0
            box[2][0] = 0
            box[2][1] = 0
        else:
            box[0][1] = box_line[3]
            box[0][2] = box_line[4]
            box[1][0] = box_line[5]
            box[1][2] = box_line[6]
            box[2][0] = box_line[7]
            box[2][1] = box_line[8]
        self.box = box
        self.make_chains()
        self.make_residues()
        self.unity = 'nm'
        return self

    def assign_moltype(self):
        """Identifies what type of molecule the Model is: protein, dna, or rna.
        If it is a mix, or if it is an organic molecule, "unknown" is
        assigned to self.moltype.
        """
        residues = set([r.resname for r in self.residues])

        # do not consider water and ions
        residues -= library._water
        residues -= library._ions

        # determine type
        # if 'residues' is not empty AND it's a subset of 'library._protein_residues'
        if bool(residues) and residues <= library._protein_residues:
            self.moltype = 'protein'
        elif bool(residues) and residues <= library._dna_residues:
            self.moltype = 'dna'
        elif bool(residues) and residues <= library._rna_residues:
            self.moltype = 'rna'
        else:
            self.moltype = 'unknown'

    def read(self, filename, bPDBTER=False, bNoNewID=True, bPDBGAP=False, bPDBMASS=False):
        """PDB/GRO file reader.

        Parameters
        ----------
        filename : str
            name of input file
        bPDBTER : bool, optional
            whether the file contains TER records?. Default is False.
        bNoNewID : bool, optional
            whether to assign new chain IDs. If True, new chain IDs starting
            with 'pmx' will be assigned(?). Only relevant if bPDBTER is also
            True. Default is True.
        bPDBGAP : bool
            whether search for gaps in the chain to assign new chain IDs.
        """
        ext = filename.split('.')[-1]
        if ext == 'pdb':
            if bPDBTER is True:
                return self.__readPDBTER(fname=filename,
                                         pdbline=None,
                                         bNoNewID=bNoNewID, bPDBGAP=bPDBGAP, bPDBMASS=bPDBMASS)
            else:
                return self.__readPDB(fname=filename)
        elif ext == 'gro':
            return self.__readGRO(filename)
        else:
            raise IOError('ERROR: Can only read pdb or gro!')

    def renumber_residues(self):
        """Renumbers all residues from 1."""
        for i, res in enumerate(self.residues):
            res.set_orig_resid(res.id)
            res.set_resid(i+1)

    # TODO/FIXME: should add/remove/append atoms all be only once in
    # Atomselection? At the moment they are repeated in Molecule, Chain,
    # and Model
    def remove_atom(self, atom):
        """Removes an Atom instance.

        Parameters
        ----------
        atom : Atom
            Atom instance to remove
        """
        m = atom.molecule
        m.remove_atom(atom)

    def remove_residue(self, residue, renumber_atoms=True, renumber_residues=True):
        """Removes a Molecule/residue instance.

        Parameters
        ----------
        residue : Molecule
            Molecule instance to remove
        renumber_atoms : bool, optional
            whether to renumber the atoms of the Model after removing the
            residue. Default is True.
        renumber_residues : bool, optional
            whether to renumber the residues of the Model after removing the
            residue. Default is True.
        """
        ch = residue.chain
        ch.remove_residue(residue, renumber_atoms=renumber_atoms,
                          renumber_residues=renumber_residues)

    def remove_chain(self, key):
        """Removes a Chain instance given the chain ID.

        Parameters
        ----------
        key : str
            ID of the chain to remove
        """
        if key not in self.chdic:
            print('No chain %s to remove....' % key)
            print('No changes applied.')
            return
        for ch in self.chains:
            if ch.id == key:
                idx = self.chains.index(ch)
                del self.chains[idx]
                del self.chdic[key]
        self.resl_from_chains()
        self.al_from_resl()
        self.renumber_residues()
        self.renumber_atoms()

    def __delitem__(self, key):
        self.remove_chain(key)

    def insert_residue(self, pos, res, chain_id):
        """Inserts a residue in Model.

        Parameters
        ----------
        pos : int
            position/index where to insert the residue
        res : Molecule
            Molecule instance to insert containig the residue of interest
        chain_id : str
            ID of the chain where to insert the residue
        """
        ch = self.chdic[chain_id]
        ch.insert_residue(pos, res)

    def replace_residue(self, residue, new, bKeepResNum=False):
        """Replaces a residue.

        Parameters
        ----------
        residue : Molecule
            residue to replace
        new : Molecule
            residue to insert
        bKeepResNum : bool, optional
            whether to keep residue ID of the residue that is inserted.
            Default is False
        """
        ch = residue.chain
        ch.replace_residue(residue, new, bKeepResNum)

    def insert_chain(self, pos, new_chain):
        """Inserts a Chain in Model.

        Parameters
        ----------
        pos : int
            index where to insert the chain within the list of chains
        new_chain : Chain
            instance of the Chain to insert in Model
        """
        if new_chain.id in self.chdic:
            print('Chain identifier %s already in use!' % new_chain.id)
            print('Changing chain identifier to 0')
            new_chain.set_chain_id('0')
        self.chains.insert(pos, new_chain)
        self.resl_from_chains()
        self.al_from_resl()
        self.make_chains()
        self.make_residues()
        self.renumber_atoms()
        self.renumber_residues()

    def append(self, new_chain):
        """ we assume chain is a Chain"""
        idx = len(self.chains)
        self.insert_chain(idx, new_chain)

    def fetch_residue(self, idx, chain=None):
        """Get a residue based on its index, or index and chain.

        Parameters
        ----------
        idx : int
            ID of the residue to fetch.
        chain : str, optional
            chain ID of the residue. This is needed when you have multiple
            chains and you have not renumbered the residues.

        Returns
        -------
        residue : Molecule
            Molecule instance of the residue found.
        """

        #########################
        # generate some residue id lists
        valid_resids = []
        for r in self.residues:
            if isinstance(r.id,str):
                valid_resids.append(r.id.replace(" ",""))
            else:
                valid_resids.append(r.id)

        if chain is not None:        
            valid_chresids = []
            for r in self.chdic[chain].residues:
                if isinstance(r.id,str):
                    valid_chresids.append(r.id.replace(" ",""))
                else:
                    valid_chresids.append(r.id)
        ##########################


        # check idx is a valid selection
        if idx not in valid_resids:#[r.id for r in self.residues]:
            raise ValueError('resid %s not found in Model residues' % idx)

        if chain is None:
            # check selection is unique
            if [r.id for r in self.residues].count(idx) > 1:
                raise ValueError('idx choice %s results in non-unique selection' % idx)
        else:
            # check chain is a valid selection
            if chain not in [c.id for c in self.chains]:
                raise ValueError('chain ID "%s" not found in Model chains' % chain)
            # check idx+chain is a valid selection
            if idx not in valid_chresids:#[r.id for r in self.chdic[chain].residues]:
                raise ValueError('resid %s not found in chain "%s"' % (idx, chain))
            # check selection is unique
            if [r.id for r in self.chdic[chain].residues].count(idx) > 1:
                raise ValueError('idx choice %s for chain "%s" results in non-unique selection' % (idx, chain))

        # then find and return the residue
        if chain is None:
            for r in self.residues:
                if isinstance(r.id,int):
                    if r.id == idx:
                        return r
                else:
                    if r.id.replace(" ","")==idx:
                        return r
        elif chain is not None:
            for r in self.chdic[chain].residues:
                if isinstance(r.id,int):
                    if r.id == idx:
                        return r
                else:
                    if r.id.replace(" ","")==idx:
                        return r
            

    def fetch_residues(self, key, inv=False):
        """Gets residues using a list of residue names.

        Parameters
        ----------
        key : str or list
            resname
        inv : bool
            invert selection. Default is False. If True, it finds all residues
            with resnames different from those in key.

        Returns
        -------
        residues : list
            list of Molecule instances with the residues
        """
        if not hasattr(key, "append"):
            key = [key]
        result = []
        if not inv:
            for r in self.residues:
                if r.resname in key:
                    result.append(r)
        else:
            for r in self.residues:
                if r.resname not in key:
                    result.append(r)
        return result

    def al_from_resl(self):
        self.atoms = []
        for r in self.residues:
            for atom in r.atoms:
                self.atoms.append(atom)

    def resl_from_chains(self):
        self.residues = []
        for ch in self.chains:
            for r in ch.residues:
                self.residues.append(r)

    def copy(self):
        return copy.deepcopy(self)

    def get_mol2_types(self):
        if self.atoms[0].symbol == '':
            self.get_symbol()
        for ch in self.chains:
            ch.get_mol2_types()

    def get_mol2_resname(self):
        for ch in self.chains:
            ch.get_mol2_resname()

    def get_nterms(self):
        nter = []
        for ch in self.chains:
            first = ch.residues[0]      # first residue
            if first.resname in library._one_letter.keys():
                nter.append(first)
        return nter

    def get_cterms(self):
        cter = []
        for ch in self.chains:
            last = ch.residues[-1]      # last residue
            if last.resname in library._one_letter.keys():
                cter.append(last)
        return last

    def rename_atoms(self):
        for c in self.chains:
            c.rename_atoms()

    def residue(self, idx):
        return self.residues[idx-1]

    def chain(self, iden):
        return self.chdic[iden]


# ==============================================================================
#                                  Functions
# ==============================================================================
def merge_models(*args):
    '''Merges the atoms from all Model objects in the list provided. Atoms will
    be merged based on the order of models in the list.

    Parameters
    ----------
    *args :
        variable length argument containing Model objects.

    Returns
    -------
    m : Model
        new Model object containing all the Models in the list.

    Examples
    --------
    >>> newmodel = merge_models(model1, model2, model3)
    >>> newmodel = merge_models(protein, ligand, cofactor, ions)
    '''
    model = Model()

    if not all(m.unity == args[0].unity for m in args):
        raise ValueError('the Model objects provided do not have the same '
                         'units - convert units to that they are compatible')

    for m in args:
        # operate on a deep copy of the model, otherwise changes in the atoms
        # of the merged model will be reflected in the parent models too
        m_ = copy.deepcopy(m)
        for a in m_.atoms:
            model.atoms.append(a)

    model.unity = model.atoms[0].unity
    model.make_chains()
    model.make_residues()
    model.assign_moltype()
    return model


def assign_masses_to_model(model, topology=None):
    '''Assigns masses to the Model atoms given the ones present in the Topology.

    Parameters
    ----------
    model : Model
        Model object of the molecule.
    topology : Topology
        Topology object of the same molecule. When no topology provided, standard library masses are used.
    '''
    
    if topology!=None:
        for ma, ta in zip(model.atoms, topology.atoms):
            if ma.name != ta.name:
                raise ValueError('mismatch of atom names between Model and '
                                 'Topology objects provided')
            ma.m = ta.m
    else:
        for a in model.atoms:
            aname = a.name
            aname = aname.upper()
            aname = aname.translate(str.maketrans('','',digits))
            if aname not in library._atommass.keys():
                a.m = 0.0
            else:
                a.m = library._atommass[aname]

def double_box(m1, m2, r=2.5, d=1.5, bLongestAxis=False, verbose=False):
    '''Places two structures (two Model objects) into a single box.
    The box is a rectangular cuboid in which the two structures are placed in
    such a way to minimise the box volume.

    Parameters
    ----------
    m1 : Model
        first structure.
    m2 : Model
        first structure.
    r : float
        distance between the two structures (nm).
    d : float
        distance to the box wall (nm).
    bLongestAxis : bool
        whether to just place structures along the
        longest axis, rather then optimising the placement and minimising
        the volume.
    verbose : bool
        whether to print out information about the new box or not.

    Returns
    -------
    mout : Model
        new Model object where the two input structures are placed within a
        single box.
    '''

    def _translate(m, v, fact=1.0):
        for a in m.atoms:
            a.x[0] = a.x[0] + fact*v[0]
            a.x[1] = a.x[1] + fact*v[1]
            a.x[2] = a.x[2] + fact*v[2]

    def _get_mass(a):
        aname = ''.join(i for i in a.name if not i.isdigit())
        if aname.startswith('Br') or aname.startswith('BR'):
            return(library._atommass['BR'])
        elif aname.startswith('Cl') or aname.startswith('CL'):
            return(library._atommass['CL'])
        elif aname.startswith('D'):
            return(library._atommass[aname[1]])
        else:
            return(library._atommass[aname[0]])
        return(1.0)

    def _principal_axes(m):
        tensor = np.matrix([[0.0, 0.0, 0.0], [0.0, 0.0, 0.0], [0.0, 0.0, 0.0]])
        for a in m.atoms:
            if ('HOH' in a.resname) or ('SOL' in a.resname) or ('WAT' in a.resname):
                continue
            mass = _get_mass(a)
            tensor[0,0] += mass*(np.power(a.x[1], 2) + np.power(a.x[2], 2))
            tensor[1,1] += mass*(np.power(a.x[0], 2) + np.power(a.x[2], 2))
            tensor[2,2] += mass*(np.power(a.x[0], 2) + np.power(a.x[1], 2))
            tensor[0,1] -= mass*a.x[0]*a.x[1]
            tensor[1,0] = tensor[0, 1]
            tensor[0,2] -= mass*a.x[0]*a.x[2]
            tensor[2,0] = tensor[0, 2]
            tensor[1,2] -= mass*a.x[1]*a.x[2]
            tensor[2,1] = tensor[1, 2]
        evals, evecs = np.linalg.eig(tensor)
        idx = evals.argsort()[::1] # sort descending, because later the rotation matrix will be calculated as transpose
        evals = evals[idx]
        evecs = evecs[:, idx]
        # check if one axis needs to be flipped
        crossprod = np.cross(np.transpose(evecs[:,0]), np.transpose(evecs[:,1]))
        dotprod = np.dot(np.transpose(evecs[:,2]), np.transpose(crossprod))
        if dotprod < 0.0:
            evecs[:,2] *= -1.0
        rotmat = np.transpose(evecs)
        return(rotmat)

    def _rotate(atoms, R):
        for atom in atoms:
            x_old = list(map(lambda x: x, atom.x))
            for i in range(3):
                atom.x[i] = 0.0
                for j in range(3):
                    atom.x[i] += x_old[j]*R[i,j]

    def _get_com_radius(m):
        com = [0.0, 0.0, 0.0]
        radius = 0.0
        mass = 0.0

        # com
        for a in m.atoms:
            a.a2nm()
            if ('HOH' in a.resname) or ('SOL' in a.resname) or ('WAT' in a.resname):
                continue
            aname = a.name[0]
            if aname.isdigit():
                aname = a.name[1]
            amass = _get_mass(a)

            com[0] += a.x[0]*amass
            com[1] += a.x[1]*amass
            com[2] += a.x[2]*amass
            mass += amass

        com[0] /= mass
        com[1] /= mass
        com[2] /= mass

        # radius
        for a in m.atoms:
            if ('HOH' in a.resname) or ('SOL' in a.resname) or ('WAT' in a.resname):
                continue
            foo = (a.x[0]-com[0])**2 + (a.x[1]-com[1])**2 + (a.x[2]-com[2])**2
            if foo > radius:
                radius = foo
        radius = np.sqrt(radius)

        return(com, radius)

    def _get_rect_dist(m):
        minx = 9999.999
        maxx = -9999.999
        miny = 9999.999
        maxy = -9999.999
        minz = 9999.999
        maxz = -9999.999
        for a in m.atoms:
            if ('HOH' in a.resname) or ('SOL' in a.resname) or ('WAT' in a.resname):
                continue
            if a.x[0] < minx:
                minx = a.x[0]
            if a.x[0] > maxx:
                maxx = a.x[0]
            if a.x[1] < miny:
                miny = a.x[1]
            if a.x[1] > maxy:
                maxy = a.x[1]
            if a.x[2] < minz:
                minz = a.x[2]
            if a.x[2] > maxz:
                maxz = a.x[2]
        a = maxx-minx
        b = maxy-miny
        c = maxz-minz
        return (a, b, c)

    def _get_extr(m, i=0):
        minx = 9999.999
        maxx = -9999.999
        for a in m.atoms:
            if ('HOH' in a.resname) or ('SOL' in a.resname) or ('WAT' in a.resname):
                continue
            if a.x[i] < minx:
                minx = a.x[i]
            if a.x[i] > maxx:
                maxx = a.x[i]
        return(minx, maxx)

    def _translate_sticking_out(m, l, i=0):
        minx = 0.0
        maxx = 0.0
        for a in m.atoms:
            if ('HOH' in a.resname) or ('SOL' in a.resname) or ('WAT' in a.resname):
                continue
            if a.x[i] < minx:
                minx = a.x[i]
            if (a.x[i]-l) > maxx:
                maxx = a.x[i]-l
        return(-1.0*(minx+maxx))

    # ====
    # Main
    # ====

    # com and radii
    com1, rad1 = _get_com_radius(m1)
    com2, rad2 = _get_com_radius(m2)

    # remove COM from the systems
    _translate(m1, com1, fact=-1.0)
    _translate(m2, com2, fact=-1.0)

    # if needed, calculate principal axes
    if bLongestAxis:
        princRot1 = _principal_axes(m1)
        princRot2 = _principal_axes(m2)
        _rotate(m1.atoms, princRot1)
        _rotate(m2.atoms, princRot2)
        a_rect1, b_rect1, c_rect1 = _get_rect_dist(m1)
        a_rect2, b_rect2, c_rect2 = _get_rect_dist(m2)

    # estimate cube's edge for the larger structure
    a_cube = 0.0
    if rad1 > rad2:
        a_cube = 2*rad1+2*d
    else:
        a_cube = 2*rad2+2*d
    if bLongestAxis is True:
        a_rect = a_rect1 + a_rect2 + r + 2.0*d
        b_rect = b_rect1 + b_rect2 + 2.0*d
        c_rect = c_rect1 + c_rect2 + 2.0*d
        if verbose is True:
            print("Cuboid dimensions (nm): ", np.round(a_rect), np.round(b_rect), np.round(c_rect, 2))
    else:
        if verbose is True:
            print("Cube's edge (nm): ", np.round(a_cube, 2))

    # cube's diagonal
    if bLongestAxis is True:
        d_cube = np.sqrt(np.power(a_rect, 2) + np.power(b_rect, 2) + np.power(c_rect, 2))
        if verbose is True:
            print("Cuboid's diagonal (nm): ", np.round(d_cube, 2))
    else:
        d_cube = a_cube*np.sqrt(3.0)
        if verbose is True:
            print("Cube's diagonal (nm): ", np.round(d_cube, 2))

    # check if cube is enough
    if bLongestAxis is True:
        a_box = a_rect
        b_box = b_rect
        c_box = c_rect
    else:
        dist_cube = d_cube - 2.0*rad1 - 2.0*rad2 - r - 2.0*d
        a_box = a_cube
        b_box = a_cube
        c_box = a_cube
        if dist_cube > 0.0:
            if verbose is True:
                print("Having a cube with the diagonal ", np.round(d_cube, 2), "nm is good enough")
        else:
            if verbose is True:
                print("Need to extend the cube")
            d_rect = 2.0*rad1 + 2.0*rad2 + r + 2.0*d
            delta_a_cube = -a_cube + np.sqrt(d_rect**2 - 2.0*a_cube**2)
            a_box = a_cube + delta_a_cube
            if verbose is True:
                print("Rectangle's edge (nm): ", np.round(a_box, 2))

    # translate the larger structure to the middle of the box
    if rad1 > rad2:
        # translate smaller to (0,b/2,c/2) and larger to (maxx2+dist+abs(minx2),b/2,c/2)
        if bLongestAxis is True:
            _translate(m2, [0.0, b_box/2.0, c_box/2.0])
            minx1, maxx1 = _get_extr(m1, 0)
            minx2, maxx2 = _get_extr(m2, 0)
            _translate(m1, [maxx2+r+abs(minx1), b_box/2.0, c_box/2.0])
        else:
            _translate(m1, [a_box/2.0, b_box/2.0, c_box/2.0])
    else:
        if bLongestAxis is True:
            _translate(m1, [0.0, b_box/2.0, c_box/2.0])
            minx1, maxx1 = _get_extr(m1, 0)
            minx2, maxx2 = _get_extr(m2, 0)
            _translate(m2, [maxx1 + r + abs(minx2), b_box/2.0, c_box/2.0])
        else:
            _translate(m2, [a_box/2.0, b_box/2.0, c_box/2.0])

    # create an output
    mout = m1
    chainList = []
    for ch in m1.chains:
        if ch.id not in chainList:
            chainList.append(ch.id)
    chainIDstring = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    chainIDchanges = {}
    for ch in m2.chains:
        if ch.id in chainList:
            # pick a new chain ID
            bFound = False
            while bFound is False:
                foo = chainIDstring[0]
                chainIDstring = chainIDstring.lstrip(chainIDstring[0])
                if foo not in chainList:
                    bFound = True
            chainList.append(foo)
            chainIDchanges[ch.id] = foo
    for a in m2.atoms:
        chID = a.chain_id
        if chID in chainIDchanges:
            a.chain_id = chainIDchanges[chID]
    mout.atoms.extend(m2.atoms)

    # create the box
    mout.box = [[a_box,0.0,0.0], [0.0,b_box,0.0], [0.0,0.0,c_box]]

    # determine translation exactly by checking how much the atoms are sticking out
    xtransl = _translate_sticking_out(mout, a_box, 0)
    ytransl = _translate_sticking_out(mout, b_box, 1)
    ztransl = _translate_sticking_out(mout, c_box, 2)
    # increase translation by a factor 1.01 to move a bit further from walls
    _translate(mout, [1.01*xtransl, 1.01*ytransl, 1.01*ztransl])

    # volume
    V = a_box*b_box*c_box
    if verbose is True:
        print("Volume: ", np.round(V, 2), "nm^3")

    return mout



#!/usr/bin/env python


import argparse
import sys
import os
from pmx import ligand_alchemy
import glob


class PMXAssembleSystem:
    """
    Entrypoint for Icolos workflow manager, builds the system and merges topologies
    """

    def __init__(
        self,
        ligandPath: str,
        edges: str,
        workPath: str,
        ff: str = "amber99sb-star-ildn-mut.ff",
        water: str = "tip3p",
    ) -> None:
        super().__init__()
        self.ligandPath = ligandPath
        self.edges: list = edges.split(" ")
        self.workPath = workPath
        self.proteinPath = os.path.join(workPath, "input/protein/")
        self.protein = self._read_protein()
        self.ff = ff
        self.water = water

    def _read_protein(self):
        protein = {}
        # read protein folder
        protein["path"] = os.path.abspath(self.proteinPath)
        # get folder contents
        protein["posre"] = []
        protein["itp"] = []
        protein["mols"] = []  # mols to add to .top
        protein["str"] = ""
        for l in glob.glob("{0}/*".format(self.proteinPath)):
            fname = l.split("/")[-1]
            if ".itp" in fname:  # posre or top
                if "posre" in fname:
                    protein["posre"].append(os.path.abspath(l))
                else:
                    protein["itp"].append(os.path.abspath(l))
                    if fname.startswith("topol_"):
                        protein["mols"].append(fname[6:-4])
                    else:
                        protein["mols"].append(fname[:-4])
            if ".pdb" in fname:
                protein["str"] = fname

        protein["mols"].sort()
        return protein

    def assemble_systems(self):
        edges = self.edges
        for edge in edges:
            parts = edge.split("_")
            lig1 = parts[0]
            lig2 = parts[1]
            lig1path = "{0}/{1}".format(self.ligandPath, lig1)
            lig2path = "{0}/{1}".format(self.ligandPath, lig2)
            hybridStrTopPath = self._get_specific_path(
                workPath=self.workPath, edge=edge, bHybridStrTop=True
            )
            outLigPath = self._get_specific_path(
                workPath=self.workPath, edge=edge, wp="unbound"
            )
            outProtPath = self._get_specific_path(
                workPath=self.workPath, edge=edge, wp="bound"
            )

            # Ligand structure
            self._make_clean_pdb(
                "{0}/mergedA.pdb".format(hybridStrTopPath),
                "{0}/init.pdb".format(outLigPath),
            )

            # Ligand+Protein structure
            self._make_clean_pdb(
                "{0}/{1}".format(self.proteinPath, self.protein["str"]),
                "{0}/init.pdb".format(outProtPath),
            )
            self._make_clean_pdb(
                "{0}/mergedA.pdb".format(hybridStrTopPath),
                "{0}/init.pdb".format(outProtPath),
                bAppend=True,
            )

            # Ligand topology
            # ffitp; these are just the atomtypes for the forcefield.
            ffitpOut = "{0}/ffmerged.itp".format(hybridStrTopPath)
            ffitpIn1 = "{0}/ffMOL.itp".format(lig1path)
            ffitpIn2 = "{0}/ffMOL.itp".format(lig2path)
            ffitpIn3 = "{0}/ffmerged.itp".format(hybridStrTopPath)
            ligand_alchemy._merge_FF_files(
                ffitpOut, ffsIn=[ffitpIn1, ffitpIn2, ffitpIn3]
            )
            # top
            ligTopFname = "{0}/topol.top".format(outLigPath)
            ligFFitp = "{0}/ffmerged.itp".format(hybridStrTopPath)
            ligItp = "{0}/merged.itp".format(hybridStrTopPath)
            itps = [ligFFitp, ligItp]
            systemName = "ligand in water"
            self.create_top(fname=ligTopFname, itp=itps, systemName=systemName)

            # Ligand+Protein topology
            # top
            protTopFname = "{0}/topol.top".format(outProtPath)
            mols = []
            for m in self.protein["mols"]:
                mols.append([m, 1])
            mols.append(["MOL", 1])
            # require the protein to be parametrised with separate itp file, not included in the .top or it will be missed.
            itps.extend(self.protein["itp"])
            systemName = "protein and ligand in water"
            self.create_top(
                fname=protTopFname, itp=itps, mols=mols, systemName=systemName
            )

    def _get_specific_path(
        self,
        workPath=None,
        edge=None,
        bHybridStrTop=False,
        wp=None,
        state=None,
        r=None,
        sim=None,
    ):
        if edge == None:
            return workPath
        edgepath = "{0}/{1}".format(workPath, edge)

        if bHybridStrTop == True:
            hybridStrPath = "{0}/hybridStrTop".format(edgepath)
            return hybridStrPath

        if wp == None:
            return edgepath
        wppath = "{0}/{1}".format(edgepath, wp)

        if state == None:
            return wppath
        statepath = "{0}/{1}".format(wppath, state)

        if r == None:
            return statepath
        runpath = "{0}/run{1}".format(statepath, r)

        if sim == None:
            return runpath
        simpath = "{0}/{1}".format(runpath, sim)
        return simpath

    def create_top(
        self,
        fname="topol.top",
        itp=["merged.itp"],
        mols=[["MOL", 1]],
        systemName="simulation system",
        destination="",
        toppaths=[],
    ):

        fp = open(fname, "w")
        # ff itp
        fp.write('#include "%s/forcefield.itp"\n' % self.ff)
        # additional itp
        for i in itp:
            fp.write('#include "%s"\n' % i)
        # water itp
        fp.write('#include "%s/%s.itp"\n' % (self.ff, self.water))
        # ions
        fp.write('#include "%s/ions.itp"\n\n' % self.ff)
        # system
        fp.write("[ system ]\n")
        fp.write("{0}\n\n".format(systemName))
        # molecules
        fp.write("[ molecules ]\n")
        for mol in mols:
            fp.write("%s %s\n" % (mol[0], mol[1]))
        fp.close()

    def _clean_backup_files(self, path):
        toclean = glob.glob("{0}/*#".format(path))
        for clean in toclean:
            os.remove(clean)

    def _make_clean_pdb(self, fnameIn, fnameOut, bAppend=False):
        # read
        fp = open(fnameIn, "r")
        lines = fp.readlines()
        out = []
        for l in lines:
            if l.startswith("ATOM") or l.startswith("HETATM"):
                out.append(l)
        fp.close()

        # write
        if bAppend == True:
            fp = open(fnameOut, "a")
        else:
            fp = open(fnameOut, "w")
        for l in out:
            fp.write(l)
        fp.close()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Entrypoint for system preparation")
    parser.add_argument("-edges")
    parser.add_argument("-ligand_path")
    parser.add_argument("-workPath")

    args = parser.parse_args()

    assembler = PMXAssembleSystem(args.ligand_path, args.edges, args.workPath)
    assembler.assemble_systems()

    sys.exit(0)



# pmx  Copyright Notice
# ============================
#
# The pmx source code is copyrighted, but you can freely use and
# copy it as long as you don't change or remove any of the copyright
# notices.
#
# ----------------------------------------------------------------------
# pmx is Copyright (C) 2006-2013 by Daniel Seeliger
#
#                        All Rights Reserved
#
# Permission to use, copy, modify, distribute, and distribute modified
# versions of this software and its documentation for any purpose and
# without fee is hereby granted, provided that the above copyright
# notice appear in all copies and that both the copyright notice and
# this permission notice appear in supporting documentation, and that
# the name of Daniel Seeliger not be used in advertising or publicity
# pertaining to distribution of the software without specific, written
# prior permission.
#
# DANIEL SEELIGER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS
# SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND
# FITNESS.  IN NO EVENT SHALL DANIEL SEELIGER BE LIABLE FOR ANY
# SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
# RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF
# CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
# ----------------------------------------------------------------------
import sys,os,shutil
import commands
from glob import glob
from pmx import *
from pmx.ndx import *
from pmx.library import _one_letter
from pmx.forcefield import MDP


ff_path = os.path.join(os.environ.get('GMXDATA'),'top/amber99sbmut.ff')


def run_command( func, string ):
    s = func.__name__+'(): '+ string
    err_file = func.__name__+'_ERROR.log'
    status, out = commands.getstatusoutput( string )
    if status != 0:
        print >>sys.stderr, 'ERROR in %s ' % func.__name__
        print >>sys.stderr, 'Output written to %s'  % os.path.abspath(os.path.join('.',err_file))
        fp = open(err_file,'w')
        print >>fp, s
        print >>fp, out
        fp.close()
        sys.exit(1)
    else:
        print "%-90s" % s, ': ok'

chain_ids = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'

class MD:

    def __init__(self, **kwargs):

        self.ff = 'amber99sbmut'
        self.water = 'tip3p' 
        self.conc = .15
        self.box_type = 'triclinic'
        self.box_size = 1.2
        self.vsite = False
        self.princ = False
        self.md_in_conf = 'md_in.pdb' 

        for key, val in kwargs.items():
            setattr(self,key,val)

    def setup(self):
        self.generate_topology( self.pdb_in)
        self.renumber_pdb( )
        self.generate_topology( self.pdb_in)
        self.generate_sim_box( )
        self.fill_sim_box(  )
        self.tpr_from_box()
        self.add_ions(  )
        self.make_em_tpr()
        self.run_em()
        self.pdb_from_em()
        self.generate_topology('min.pdb')
        self.compact_repr( 'min.pdb', self.md_in_conf)
        self.clean_backups()

    def write_residue_map( self, model ):
        fp = open('residue_map.txt','w')
        for r in model.residues:
            if r.chain_id == ' ':
                chain = '_'
                r.old_chain_id = '_'
            else:
                chain = r.chain_id
            print >>fp, '%d|%s|%s -> %d|%s|%s' %( r.orig_id, r.resname, r.old_chain_id, r.id, r.resname, chain)
        fp.close()

    def renumber_pdb(self ):
        m = Model('gmx.pdb')
        for i, chain in enumerate(m.chains):
            if chain.id != ' ':
                for r in chain.residues:
                    r.old_chain_id = chain.id
                chain.set_chain_id( chain_ids[i] )
        self.write_residue_map( m )
        m.write('start.pdb')
        self.pdb_in = 'start.pdb'
        
    def __str__(self):
        s = '< MD (%s) > ' % self.pdb_in
        return s

    def clean_backups(self):
        files = glob('#*#')
        for f in files:
            os.unlink(f)
    
    def generate_topology(self, pdb_in):
    
        run = 'pdb2gmx -f %s -water %s -ff %s -o gmx.pdb ' % ( pdb_in, self.water, self.ff )
        if self.vsite:
            run+=' -vsite hydrogens'
        run_command( self.generate_topology, run )

    def generate_sim_box(self ):
        if self.princ:
            run = 'echo 4 | editconf -f gmx.pdb -o tmp.pdb -c -princ '
            run_command( run_editconf, run)
            run = 'editconf -f tmp.pdb -o ed.pdb -d %g -bt %s ' % (self.box_size, self.box_type)
            run_command( generate_sim_box, run)
        else:
            run = 'editconf -f gmx.pdb -o ed.pdb -d %g -bt %s ' % (self.box_size, self.box_type)
            run_command( self.generate_sim_box, run)
    
    def fill_sim_box(self ):
        if self.water == 'spce': water = 'spc216'
        else: water = self.water
        run = 'genbox -cp ed.pdb -cs %s -p -o box.pdb' % os.path.join(ff_path,water)
        run_command( self.fill_sim_box, run)
    
    def tpr_from_box(self):
        run = 'grompp -f ~/mdp/em.mdp -c box.pdb'
        run_command( self.tpr_from_box, run)

    def get_solvent_index(self ):
        run = 'echo q | make_ndx -f topol.tpr -o tmp.ndx'
        run_command( self.get_solvent_index, run)
        ndx = IndexFile("tmp.ndx")
        return ndx.names.index('SOL')
    
    def add_ions(self ):
        idx = self.get_solvent_index()
        run = 'echo %d | genion  -conc %g -neutral -p -o ion.pdb' % (idx, self.conc)
        run_command( self.add_ions, run)
    
    def make_em_tpr(self):
        run = 'grompp -f ~/mdp/em.mdp -c ion.pdb'
        run_command( self.make_em_tpr, run)
    
    def run_em(self):
        run = 'mdrun -v -c em.pdb'
        run_command( self.run_em, run)
    
    def pdb_from_em(self):
        run = 'echo 0| trjconv -f em.pdb -s topol.tpr -o min.pdb' 
        run_command( self.pdb_from_em, run)
    
    def compact_repr(self, pdb_in, pdb_out ):
        run = 'echo 0 | trjconv -f %s -s topol.tpr -ur compact -pbc mol -o %s' % (pdb_in, pdb_out)
        run_command( self.compact_repr, run)
    


class FreeEnergyMD(MD):

    def __init__(self, mutation_file, **kwargs):

        MD.__init__( self )
        
        for key, val in kwargs.items():
            setattr(self,key,val)
        self.mutation_tags = []
        self.read_mutation_file( mutation_file )
        self.mutations = []
        if self.mutation_tags:
            self.mutations_from_tags()
        self.runs = []
        self.is_single_chain = False
        
    def read_mutation_file(self, mut_file ):
        print '\n\t\t\tReading mutation file: %s\n' % mut_file
        l = open(mut_file).readlines()
        count = 1
        for line in l:
            entr = line.strip()
            if entr:
                print '\t\t\t (%d) ->  %s' % (count, entr)
                self.mutation_tags.append( entr )
                count+=1
        

    def setup(self):
        self.read_pdb()
        print '\n\n'
        for m in self.mutations:
            self.setup_mutation_run(m)

    def read_pdb( self ):
        self.model = Model(self.md_in_conf)
        if len(self.model.chains) == 1:
            self.is_single_chain = True
            
    def mutations_from_tags(self):
        for t in self.mutation_tags:
            mut = t.split()
            new_mut = []
            for m in mut:
                resid, resn, chain = m.split('|')
                resid = int(resid)
                new_mut.append( ( resid, resn, chain ) )
            self.mutations.append( new_mut )
    
    def setup_mutation_run(self, mutation ):
        muts = []
        affected_chains = []
        for m in mutation:
            resid, resn, chain = m
            residue = self.model.residues[resid-1]
            resname = _one_letter[residue.resname]
            muts.append( chain+'.'+resname+str(resid)+resn )
            if chain not in affected_chains:
                affected_chains.append( chain )
        name = '-'.join(muts)
        if str(self.__class__).split('.')[1] == 'DiscreteTI':
            name+='.dti'
        elif str(self.__class__).split('.')[1] == 'CrooksMD':
            name+='.crooks'
        print '\n\t\t\tPreparing mutation run -> %s \n' % name
        if os.path.isdir( name ):
            shutil.rmtree(name)
        os.mkdir(name)
        self.runs.append( name )
        shutil.copy(self.md_in_conf, name)
        script_file = os.path.join(name,'mutations.txt')
        fp = open(script_file,'w')
        for m in mutation:
            resid, resn, chain = m
            print >>fp, resid, resn
        fp.close()
        self.make_mutation(name, affected_chains)

        
    def make_mutation(self, path, affected_chains ):
        os.chdir(path)
        run = '~/software/pmx/scripts/mutate_beta45.py -f %s -script mutations.txt -o mut.pdb' % self.md_in_conf
        run_command( self.make_mutation, run )
        self.generate_topology( 'mut.pdb')
        if self.is_single_chain:
            itp_file = 'topol_Protein.itp' 
            run = '~/software/pmx/scripts/make_bstate_beta45.py -itp %s' % itp_file
            run_command( self.make_mutation, run )
            shutil.move(itp_file, itp_file+'.save')
            shutil.move('newtop.itp', itp_file)
        else:
            for chain in affected_chains:
                print 'Applying changes to topology of chain %s' % chain
                itp_file = 'topol_Protein_chain_%s.itp' % chain
                run = '~/software/pmx/scripts/make_bstate_beta45.py -itp %s' % itp_file
                run_command( self.make_mutation, run )
                shutil.move(itp_file, itp_file+'.save')
                shutil.move('newtop.itp', itp_file)
        os.chdir('..')            
        

class CrooksMD( FreeEnergyMD ):

    def __init__(self, mutation_file, **kwargs):
        FreeEnergyMD.__init__(self, mutation_file)
        for key, val in kwargs.items():
            setattr(self,key,val)


    def do_crooks( self, mdp_file, min_mdp_file, time, nruns ):
        mdp = MDP().read( mdp_file)
        min_mdp = MDP().read( min_mdp_file)
        self.prepare_min_mdp_files(min_mdp, time )
        self.prepare_eq_mdp_files(mdp, time )
        for r in self.runs:
            print '\n\t\t\tSetting up Crooks run -> %s\n' % r

            self.minimize_states( r )
            self.setup_equilibration_runs( r, nruns )
        
    def minimize_states( self, path ):

        os.chdir(path)
        run = 'grompp -f ../crooks_minA.mdp -c gmx.pdb -o minA.tpr'
        run_command( self.minimize_states, run )
        run = 'grompp -f ../crooks_minB.mdp -c gmx.pdb -o minB.tpr'
        run_command( self.minimize_states, run )
        print '\n\t\t\tRunning energy minimization on %s ( state A ) \n' % ( path )
        run = 'mdrun -v -c emA.pdb -s minA.tpr'
        run_command( self.minimize_states, run )
        print '\n\t\t\tRunning energy minimization on %s ( state B ) \n' % ( path )
        run = 'mdrun -v -c emB.pdb -s minB.tpr'
        run_command( self.minimize_states, run )
        os.chdir('..')
        
    def prepare_eq_mdp_files( self, mdp, time ):
        nsteps = 1000*time/.002
        mdp['nsteps'] = nsteps
        mdp['init-lambda'] = 0
        fp = open('crooks_eqA.mdp','w')
        print >>fp, mdp
        fp.close()
        mdp['init-lambda'] = 1
        fp = open('crooks_eqB.mdp','w')
        print >>fp, mdp
        fp.close()

    def prepare_min_mdp_files( self, mdp, time ):
        mdp['init-lambda'] = 0
        fp = open('crooks_minA.mdp','w')
        print >>fp, mdp
        fp.close()
        mdp['init-lambda'] = 1
        fp = open('crooks_minB.mdp','w')
        print >>fp, mdp
        fp.close()

        
    def setup_equilibration_runs( self, path, nruns ):
        os.chdir(path)
        for i in range(nruns):
            print '\n\t\t\tPreparing run input file for  %s ( run %d ) \n' % ( path, i )
            os.mkdir('runA.%d' % i)
            os.mkdir('runB.%d' % i)
            run = 'grompp -f ../crooks_eqA.mdp -c emA.pdb -o runA.%d/topol.tpr' % i
            run_command( self.setup_equilibration_runs, run )
            run = 'grompp -f ../crooks_eqB.mdp -c emB.pdb -o runB.%d/topol.tpr' % i
            run_command( self.setup_equilibration_runs, run )
            self.clean_backups()
        os.chdir('..')


class DiscreteTI( FreeEnergyMD ):

    def __init__(self, mutation_file, **kwargs):
        FreeEnergyMD.__init__(self, mutation_file)
        for key, val in kwargs.items():
            setattr(self,key,val)
        
    def read_lambda_steps( self, filename ):
        l = open(filename).readlines()
        self.lambda_steps = []
        for line in l:
            entr = line.strip()
            if entr:
                self.lambda_steps.append( float(entr) )

    def prepare_dti_mdp_files( self, mdp_file, time ):
        mdp = MDP().read(mdp_file)
        nsteps = 1000*time/.002
        mdp['nsteps'] = nsteps
        for lda in self.lambda_steps:
            fp = open('dti_%4.3f.mdp' % round(lda,3),'w' )
            mdp['init-lambda'] = lda
            print >>fp, mdp
            fp.close()

    def prepare_dti_min_mdp_files( self, mdp_file):
        mdp = MDP().read(mdp_file)
        for lda in self.lambda_steps:
            fp = open('dti_min_%4.3f.mdp' % round(lda,3),'w' )
            mdp['init-lambda'] = lda
            print >>fp, mdp
            fp.close()

    def setup_runs( self, path ):
        os.chdir( path )
        for lda in self.lambda_steps:
            min_mdp = 'dti_min_%4.3f.mdp' % round(lda,3)
            run_mdp = 'dti_%4.3f.mdp' % round(lda,3)
            run_dir = 'run_%4.3f' % round(lda,3)
            os.mkdir( run_dir )
            print '\n\t\t\tRunning energy minimization on %s/%s \n' % ( path, run_dir )
            run = 'grompp -f ../%s -c gmx.pdb -o %s/em.tpr' % (min_mdp, run_dir )
            run_command( self.setup_runs, run )
            os.chdir( run_dir )
            run = 'mdrun -v -c em.pdb -s em.tpr'
            run_command( self.setup_runs, run )
            os.chdir('..')
            print '\n\t\t\tPreparing run input file for  %s/%s \n' % ( path, run_dir )
            run = 'grompp -f ../%s -c %s/em.pdb -o %s/topol.tpr' % (run_mdp, run_dir, run_dir )
            run_command( self.setup_runs, run )
            self.clean_backups()
        os.chdir( '..')
        
    def do_dti( self, mdp, min_mdp, time ):
        self.prepare_dti_min_mdp_files( min_mdp)
        self.prepare_dti_mdp_files( mdp, time)
        for r in self.runs:
            print '\n\t\t\tSetting up Discrete TI run -> %s\n' % r
            self.setup_runs( r )


        
def main(argv):

    version = "1.0"

    options = [
        Option( "-water", "string", "tip3p", "water model"),
        Option( "-box_type", "string", "triclinic", "box geometry (triclinic, octahedron, dodecahedron)"),
        Option( "-box_size", "float", 1.2, "distance from solute to box"),
        Option( "-conc", "float", 0.15, "ion concentration"),
        Option( "-vsite", "bool", False, "use virtual sites"),
        Option( "-fe.dti", "bool", False, "do discrete TI setup"),
        Option( "-dti_run_time", "float", 10., "Simulation time [ns] at each lambda point"),
        Option( "-fe.crooks", "bool", False, "do Crooks setup"),
        Option( "-n_crooks_runs", "int", 1, "setup # crooks runs for each mutation"),
        Option( "-crooks_run_time", "float", 50., "Simulation time [ns] for crooks equilibrium runs"),
        Option( "-skip_md_setup", "bool", False, "skip md setup and use -f as starting configuration for free energy runs"),
        ]
    
    files = [
        FileOption("-f", "r",["pdb"],"protein", "input pdb file"),
        FileOption("-m", "r/o",["txt"],"mutations", "mutations to make"),
        FileOption("-crooks_mdp", "r/o",["mdp"],"template", "template run mdp file for crooks equilibrium runs"),
        FileOption("-dti_mdp", "r/o",["mdp"],"template", "template run mdp file for discrete TI calculations"),
        FileOption("-min_mdp", "r/o",["mdp"],"em", "template minimization mdp file ( for TI or Crooks )"),
        FileOption("-lambda_steps", "r/o",["txt"],"lambda_steps", "text file with lambda steps for DTI runs"),
        ]
    
    
    
    help_text = ("Script for setting up plain MD runs",
                 )

    
    cmdl = Commandline( argv, options = options,
                        fileoptions = files,
                        program_desc = help_text,
                        check_for_existing_files = False, version = version)


    # make some basic checks
    bFreeEnergy = False
    if cmdl['-fe.dti'] or cmdl['-fe.crooks']:
        bFreeEnergy = True
        mutation_file = cmdl['-m']
        try:
            open(mutation_file).readlines()
        except:
            print >>sys.stderr, 'Error: Cannot open %s' % mutation_file
            sys.exit(1)
        if cmdl['-fe.dti']: # require lambda steps
            try:
                open(cmdl['-lambda_steps']).readlines()
            except:
                print >>sys.stderr, 'Error: Cannot open %s' % cmdl['-lambda_steps']
                sys.exit(1)
            


    pdb_in = cmdl['-f']
    water = cmdl['-water']
    box_type = cmdl['-box_type']
    box_size = cmdl['-box_size']
    conc = cmdl['-conc']
    free_energy_start_pdb = None
    bVsite = cmdl['-vsite']
    if bVsite and bFreeEnergy:
        print >>sys.stderr, 'Error: Cannot use virtual sites in free energy calculations !' 
        sys.exit(1)
        

    if not cmdl['-skip_md_setup']:
        md = MD( pdb_in = pdb_in, water = water, conc = conc, box_type = box_type, box_size = box_size )
        md.setup()
        free_energy_start_pdb = md.md_in_conf
    else:
        if not bFreeEnergy:
            print 'Nothing to do........... (no MD, no Free Energy)'
            sys.exit()
    if bFreeEnergy:
        if not free_energy_start_pdb:
            free_energy_start_pdb = pdb_in
        if cmdl['-fe.crooks']:
            crooksMD = CrooksMD( mutation_file, md_in_conf = free_energy_start_pdb)
            crooksMD.setup()
            crooksMD.do_crooks( cmdl['-crooks_mdp'], cmdl['-min_mdp'], cmdl['-crooks_run_time'], cmdl['-n_crooks_runs'] )

        if cmdl['-fe.dti']:
            dti = DiscreteTI( mutation_file, md_in_conf = free_energy_start_pdb )
            dti.setup()
            dti.read_lambda_steps( cmdl['-lambda_steps'] )
            dti.do_dti(cmdl['-dti_mdp'], cmdl['-min_mdp'], cmdl['-dti_run_time'] )


    print '\n\t\t\t ....... DONE .......... \n'
    


##     n_crooks_runs = cmdl['-n_crooks_runs']
##     if cmdl.opt['-m'].is_set:
##         mut_file  = cmdl['-m']
##     else:
##         mutations = []

##     if cmdl.opt['-mdp'].is_set:
##         mdp = MDP().read( cmdl['-mdp'] )
##     if cmdl.opt['-min_mdp'].is_set:
##         min_mdp = MDP().read( cmdl['-min_mdp'] )
##     if cmdl.opt['-lambda_steps'].is_set:
##         lambda_file = cmdl['-lambda_steps'] 
    
        
##     md = MD( pdb_in = pdb_in, water = water, conc = conc, box_type = box_type, box_size = box_size )
## #    md.setup()

##     dti = DiscreteTI( mut_file, md_in_conf = md.md_in_conf )
##     dti.setup()
##     dti.read_lambda_steps( lambda_file )
##     dti.do_dti(mdp, min_mdp, 10. )


    
##     crooksMD = CrooksMD( mut_file, md_in_conf = md.md_in_conf)
##     crooksMD.setup()
##     crooksMD.do_crooks(mdp, 20, cmdl['-n_crooks_runs'])
    
#    fe_md = FreeEnergyMD( mutations, md_in_conf = md.md_in_conf )
#    fe_md.setup()
    
    
##     generate_topology(pdb_in, water, 'amber99sbmut')
##     generate_sim_box( box_type, box_size )
##     fill_sim_box( water )
##     tpr_from_box()
##     add_ions( conc )
##     make_em_tpr()
##     run_em()
##     pdb_from_em()
##     generate_topology('min.pdb', water, 'amber99sbmut')
##     compact_repr( 'min.pdb')
##     clean_backups()



if __name__=='__main__':
    main( sys.argv )
    




#! /usr/bin/env python

import argparse
import os
import shutil
import numpy as np
from pmx.model import Model, merge_models, assign_masses_to_model, double_box
from pmx.alchemy import AbsRestraints, couple_mol, decouple_mol
from pmx.forcefield import Topology, merge_atomtypes
from pmx import gmx
from copy import deepcopy
from time import sleep
from pmx.scripts.cli import check_unknown_cmd


# TODO: build folder structure and mdp files for equil or nonequil calcs?


def parse_options():
    parser = argparse.ArgumentParser(description='''
This scripts helps to setup an absolute binding free energy calculation.
As a minimal input, you need to provide a structure and topology file for both
the protein (or host) and ligand (or guest) molecule.

The topology is setup so to contain restraints as defined by Boresch et al. (2003)
J Phys Chem B 107(35); these include one distance, two angles, and three dihedrals
between ligand and protein. You can either provide explicitly the atoms to be
included in the restraints, or let the script choose them automatically.
''')
    parser.add_argument('-pt',
                        metavar='protop',
                        dest='pro_top',
                        type=str,
                        help='Input topology file for the protein. '
                        'Default is "protein.top".',
                        default='protein.top')
    parser.add_argument('-lt',
                        metavar='ligtop',
                        dest='lig_top',
                        type=str,
                        help='Input topology file for the ligand. It is '
                        'expected that all params needed for the ligand '
                        'are explicitly defined in this file. '
                        'Default is "ligand.itp".',
                        default='ligand.itp')
    parser.add_argument('-pc',
                        metavar='procrd',
                        dest='pro_crd',
                        type=str,
                        help='Input structure file in PDB or GRO format '
                        'for the protein. Default is "protein.gro".',
                        default='protein.gro')
    parser.add_argument('-lc',
                        metavar='ligcrd',
                        dest='lig_crd',
                        type=str,
                        help='Input structure file in PDB or GRO format '
                        'for the ligand. Default is "ligand.gro".',
                        default='ligand.gro')
    parser.add_argument('--build',
                        dest='build',
                        help='Whether to build the system (editconf, solvate, '
                        'genion) with a standard setup once the input files '
                        '(top, gro) are ready.',
                        default=False,
                        action='store_true')
    parser.add_argument('--doublebox',
                        dest='doublebox',
                        help='Whether to use the double-system single-box '
                        'setup. This is useful for charged ligands. '
                        'Default is False.',
                        default=False,
                        action='store_true')
    parser.add_argument('--longest_axis',
                        dest='longest_axis',
                        help='Whether to just place structures along the '
                        'longest axis, rather then minimising the volume. '
                        'This option is relevant only when using '
                        '--doublebox. Default is False.',
                        default=False,
                        action='store_true')
    parser.add_argument('--keep_intra',
                        dest='keep_intra',
                        help='Whether to keep the LJ intramolecular '
                        'interactions when the ligand is decoupled. '
                        'This option is relevant only when using '
                        '--doublebox. Default is False.',
                        default=False,
                        action='store_true')
    parser.add_argument('--lig_ids',
                        metavar='',
                        dest='lig_ids',
                        help='Three atom indices. If provided, these will be '
                        'used for the protein-ligand restraints. Otherwise '
                        'they are chosen automatically.',
                        default=None,
                        nargs=3)
    parser.add_argument('--pro_ids',
                        metavar='',
                        dest='pro_ids',
                        help='Three atom indices. If provided, these will be '
                        'used for the protein-ligand restraints. Otherwise '
                        'they are chosen automatically.',
                        default=None,
                        nargs=3)
    parser.add_argument('--restr_switch_on',
                        dest='restr_switch_on',
                        help='Whether to switch the restraints on or off, '
                        'where "on" means no restraints in stateA, and "off" '
                        'means no restraints in state B. '
                        'Default is True (switch on).',
                        default=True,
                        action='store_false')
    parser.add_argument('--seed',
                        metavar='int',
                        dest='seed',
                        help='Random seed to use when picking atoms for the '
                        'restraints. The automated restraints selection is '
                        'stochastic, so if you want to have a reproducible '
                        'behaviour, provide a random seed.',
                        default=None,
                        type=int)

    args, unknown = parser.parse_known_args()
    check_unknown_cmd(unknown)

    return args


def main(args):

    # Import GRO and TOP files
    lig = Model(args.lig_crd, renumber_residues=False)
    pro = Model(args.pro_crd, renumber_residues=False)
    ligtop = Topology(args.lig_top, is_itp=True, assign_types=True, self_contained=True)
    protop = Topology(args.pro_top, assign_types=False)

    # check lig has only 1 residue
    assert len(lig.residues) == 1

    # assign masses to ligand atoms
    assign_masses_to_model(lig, ligtop)

    # add position restraints to ligand topology if missing
    if not ligtop.footer and ligtop.has_posre is False:
        ligtop.make_posre()

    # ----------------------
    # merge gro into complex
    # ----------------------
    com = merge_models(lig, pro)
    com.box = pro.box
    com.renumber_atoms()

    # ------------------------
    # make topology of complex
    # ------------------------
    comtop = deepcopy(protop)
    comtop.include_itps.append(("ligand.itp", "top"))
    comtop.system = 'Complex'
    comtop.molecules.insert(0, [ligtop.name, 1])

    # merge atomtypes
    comtop.atomtypes = merge_atomtypes(ligtop.atomtypes, protop.atomtypes)

    # ------------------------------------------
    # figure out restraints (and save top entry)
    # ------------------------------------------
    restraints = AbsRestraints(protein=pro, ligand=lig,
                               pro_ids=args.pro_ids, lig_ids=args.lig_ids,
                               seed=args.seed)

    # write restraints info
    restraints.write_summary()

    if args.build is False:
        # write restraints section only if we are not setting up the system
        # solvate/genion do not work well with topologies that contain
        # this restraints section
        comtop.ii = restraints.make_ii(switch_on=args.restr_switch_on,
                                       ligand_first=True)

    # -------------------------------
    # save coordinates/topology files
    # -------------------------------
    com.write('complex.gro')
    ligtop.write('ligand.itp', stateBonded='A', write_atypes=False, posre_include=True)
    comtop.write('complex.top', stateBonded='A')

    # ---------------------------------------------------
    # gmx setup: editconf, solvate, genion, mdp files etc
    # ---------------------------------------------------
    if args.build is True:
        goodtogo = _check_topology_has_all_ff_info(protop)
        if goodtogo is False:
            print('Skipping system building ...')
            sleep(3)

        # ================
        # single-box setup
        # ================
        if args.doublebox is True and goodtogo is True:

            # create double-system single-box
            # the second ligand (lig) is inserted before the complex in the
            # output gro file (easier to modify topology this way atm)
            mout = double_box(m1=com, m2=lig, r=2.5, d=1.5,
                              bLongestAxis=args.longest_axis, verbose=False)
            mout.write('doublebox.gro')

            # identify an atom to position restrain: this will be used in both
            # ligands to keep ligand and complex apart
            # reload because indices have changed after lig was merged into complex
            lig = Model(args.lig_crd, renumber_residues=False)
            assign_masses_to_model(lig, ligtop)
            posre_idx = _find_atom_near_com(lig)

            # create ligand topology with B-state
            # ------------------------------------
            ligtopAB = deepcopy(ligtop)
            decouple_mol(ligtopAB)
            _add_fixed_posre(top=ligtopAB, n=posre_idx)
            ligtopAB.write('ligandAB.itp', stateBonded='A', stateTypes='AB', stateQ='AB',
                           write_atypes=False, posre_include=True)

            # create second ligand topology with B-state
            # ------------------------------------------
            ligtopBA = deepcopy(ligtop)
            couple_mol(ligtopBA)
            _add_fixed_posre(top=ligtopBA, n=posre_idx)
            ligtopBA.name = '{}2'.format(ligtopBA.name)
            ligtopBA.write('ligandBA.itp', stateBonded='A', stateTypes='AB', stateQ='AB',
                           write_atypes=False, posre_include=True)

            # create system topology
            # ----------------------
            doubletop = deepcopy(comtop)
            doubletop.include_itps = [('ligandAB.itp', 'top'), ('ligandBA.itp', 'bottom')]
            doubletop.molecules.append([ligtopBA.name, 1])
            # merge atomtypes (basically add the dummies)
            doubletop.atomtypes = merge_atomtypes(doubletop.atomtypes, ligtopAB.atomtypes)
            # keep intramolecular interactions?
            if args.keep_intra is True:
                doubletop.make_nonbond_params(rule=2)

            # save topology
            doubletop.write('doublebox.top', stateBonded='A')

            # run gromacs setup
            # -----------------
            gmx.solvate(cp='doublebox.gro', cs='spc216.gro', p='doublebox.top', o='solvate.gro')
            gmx.write_mdp(mdp='enmin', fout='genion.mdp')
            gmx.grompp(f='genion.mdp', c='solvate.gro', p='doublebox.top', o='genion.tpr', maxwarn=1)
            gmx.genion(s='genion.tpr', p='doublebox.top', o='genion.gro', conc=0.15, neutral=True)

            # add restraints to topology
            doubletop = Topology('doublebox.top', assign_types=False)
            doubletop.ii = restraints.make_ii(switch_on=args.restr_switch_on,
                                              ligand_first=True)
            doubletop.write('doublebox.top', stateBonded='A')

        # ==================================
        # standard setup with separate boxes
        # ==================================
        elif args.doublebox is False and goodtogo is True:

            # Setup complex
            # -------------
            os.mkdir('complex')
            os.chdir('complex')

            shutil.copy('../complex.gro', '.')
            shutil.copy('../complex.top', '.')
            shutil.copy('../ligand.itp', '.')

            gmx.editconf(f='complex.gro', o='editconf.gro', bt='cubic', d=1.2)
            gmx.solvate(cp='editconf.gro', cs='spc216.gro', p='complex.top', o='solvate.gro')
            gmx.write_mdp(mdp='enmin', fout='genion.mdp')
            gmx.grompp(f='genion.mdp', c='solvate.gro', p='complex.top', o='genion.tpr', maxwarn=1)
            gmx.genion(s='genion.tpr', p='complex.top', o='genion.gro', conc=0.15, neutral=True)

            # add restraints to topology
            comtop = Topology('complex.top', assign_types=False)
            comtop.ii = restraints.make_ii(switch_on=args.restr_switch_on,
                                           ligand_first=True)
            comtop.write('complex.top', stateBonded='A')

            os.chdir('../')

            # Setup ligand
            # ------------
            os.mkdir('ligand')
            os.chdir('ligand')

            lig.write('ligand.gro')
            # setup topology
            # use the same protein ff for the water/ions in ligand sims
            ligtop.is_itp = False  # now we want to write it as top file
            ligtop.forcefield = protop.forcefield
            ligtop.footer = ['#include "{ff}.ff/tip3p.itp"'.format(ff=ligtop.forcefield),
                             '#ifdef POSRES_WATER',
                             '[ position_restraints ]',
                             '1    1       1000       1000       1000',
                             '#endif',
                             '#include "{ff}.ff/ions.itp"'.format(ff=ligtop.forcefield)]

            # add system and molecules
            ligtop.system = 'ligand'
            ligtop.molecules = [[ligtop.name, 1]]
            ligtop.write('ligand.top', stateBonded='A', write_atypes=True,
                         posre_include=True)

            # run gromacs setup
            gmx.editconf(f='ligand.gro', o='editconf.gro', bt='cubic', d=1.2)
            gmx.solvate(cp='editconf.gro', cs='spc216.gro', p='ligand.top', o='solvate.gro')
            gmx.write_mdp(mdp='enmin', fout='genion.mdp')
            gmx.grompp(f='genion.mdp', c='solvate.gro', p='ligand.top', o='genion.tpr', maxwarn=1)
            gmx.genion(s='genion.tpr', p='ligand.top', o='genion.gro', conc=0.15, neutral=True)

            os.chdir('../')


    print('\n\n          ********** Setup Completed **********\n\n')
    if args.build is True and args.doublebox is False and goodtogo is True:
        print('The input files for the simulations of the complex are:')
        print('    complex/complex.top')
        print('    complex/genion.gro')
        print('')
        print('The input files for the simulations of the ligand are:')
        print('    ligand/ligand.top')
        print('    ligand/genion.gro')
        print('')
        print('Information about the restraints are in:')
        print('    restraints.info')
    elif args.build is True and args.doublebox is True and goodtogo is True:
        print('The input files for the simulations are:')
        print('    doublebox.top')
        print('    genion.gro')
        print('')
        print('Information about the restraints are in:')
        print('    restraints.info')


def _check_topology_has_all_ff_info(top):
    '''This is needed in particular for host-guest systems where the host has
    not passed through pdb2gmx, and might not contain the ff info needed for
    solvate/genion/grompp (i.e. water/ion params to use).
    '''
    good = True

    if top.forcefield == '':
        good = False
        print('\nthe input topology file {} does not seem to '
              'contain information on the forcefield '
              'parameters to use.\nThese are usually defined via the following '
              'statement at the top of the topology file:\n'
              '#include "amber99sb-ildn.ff/forcefield.itp"\n'
              'Without this information it is not possible '
              'to setup the system in Gromacs.\n'.format(top.filename))

    if 'ions.itp' not in " ".join(top.footer):
        good = False
        print('\nthe input topology file {} does not seem to '
              'contain information on the ion '
              'parameters to use.\nThese are usualy defined '
              'just above [ system ], e.g.:\n'
              '#include "amber99sb-ildn.ff/ions.itp"\n'
              'Without this information it is not possible '
              'to setup the system in Gromacs.\n'.format(top.filename))

    if 'tip' not in " ".join(top.footer) and 'spc' not in " ".join(top.footer):
        good = False
        print('\nthe input topology file {} does not seem to '
              'contain information on the water '
              'parameters to use.\nThese are usualy defined '
              'just above [ system ], e.g.:\n'
              '#include "amber99sb-ildn.ff/tip3p.itp"\n'
              'Without this information it is not possible '
              'to setup the system in Gromacs.\n'.format(top.filename))

    return good


def _add_fixed_posre(top, n):
    ''' n : index of atom to restrain
    '''
    posre = ['[ position_restraints ]',
             '   {n}      1      1000   1000   1000'.format(n=n)]

    top.footer = posre + top.footer


def _find_atom_near_com(m):
    # get com
    com = np.array(m.com(vector_only=True))
    coords = [np.array(a.x) for a in m.atoms]
    # get all distances
    distances = [np.linalg.norm(c-com) for c in coords]
    # get index of atom closest to com. Add 1 since gromacs uses idx from 1
    com_idx = np.argmin(distances)+1
    return com_idx


def entry_point():
    args = parse_options()
    main(args)


if __name__ == '__main__':
    entry_point()



#!/usr/bin/env python
import os
from pmx import model
from subprocess import call


def main():
    path = os.path.abspath(model.__file__)
    dir_path = os.path.dirname(path)
    gmxlib = os.path.join(dir_path, 'data/mutff')

    print('\n  In order to be able to use the hybrid/alchemical force fields \n'
          '  available in pmx, the environment variable GMXLIB needs to be set.\n')

    print('  The path to your pmx force field library is the following:')
    print('  %s\n' % gmxlib)

    print('  You can either set GMXLIB in your shell session as follows:')
    print('  $ export GMXLIB=%s\n' % gmxlib)
    print('  Or you can add this directly in your bashrc file.\n')

    set_gmxlib = input('  Do you wish pmx to set the GMXLIB variable in '
                       'your ~/.bashrc? [yes|no]\n  >>> ')

    while set_gmxlib not in ['yes', 'no']:
        set_gmxlib = input('  Please enter only either "yes" or "no"\n  >>> ')

    if set_gmxlib == 'yes':
        call('echo "# GMXLIB for using pmx forcefields" >> ~/.bashrc', shell=True)
        call('echo "export GMXLIB=%s" >> ~/.bashrc' % gmxlib, shell=True)
        print('  \n  All done. GMXLIB is set in your ~/.bashrc file.\n')
    elif set_gmxlib == 'no':
        print('  \n  All done. Do not forget to set GMXLIB any time you want to \n'
              '  use pmx or Gromacs with one of the hybrid forcefields in pmx.\n')


def entry_point():
    main()


if __name__ == '__main__':
    entry_point()
